
/****************************************************************************

    Level 2 library for basic UART operating system.

    This library contains follow functions,

	s_putc		A UART write function that waits for a TRUE
			transmitter status, then writes a byte to the serial
			port.

	s_getc		A UART read function that waits for a TRUE receiver
			status, then reads a byte from the serial port.

	s_inchar	A UART read function that returns a character if the
			receiver status is TRUE, otherwise returns an error.

	vsetbits	Builds a new bit pattern in the virtual register and
			call level-0 _setbits to install that bit pattern in
			the physical UART's register.

	vsetbr		Sets the baud rate to the rank element of the
			structure's divisor array. Passes high- and low-order
			divisors to the level-1 function, _vsetbr.

	vset232 	Turns selected RS-232 output ON or OFF.

	vstat232	Returns status of RS-232 input pointed to by p232.

	s_fputc 	Provides "formated" character-by-character serial
			output according to the soctl_ structure.

	s_eputc 	Outputs the byte to the SIO and, if consflg is set,
			to standard output.

	s_fgetc 	Provides "formatted" character-by-character serial
			input according to the sictl_ structure pointed to
			by the SIO.

	s_iflush	Returns pending byte from the s_octl's "look-back"
			buffer.

	xon		Macros defined in SIODEF.H for tramsmitting XON and
	xoff		XOFF.

****************************************************************************/

#include    "ascii.h"
#include    "stdio.h"
#include    "siodef.h"
#include    "sioctl.h"

/*
FUNCTION NAME:	s_getc
LEVEL:		2
LIBRARY:	BOUS.LIB
DESCRIPTION:	Wait for a byte to arrive from the serial port (SIO) pointed
		to by siop.
OTHER MODULES:	None
RETURNS:	BYTE
COMMENTS:	Access to the port is through a pointer to an sio structure.
*/

BYTE s_getc(siop)
register SIO *siop;
{
    while ((*siop->s_rstat)(siop) == 0)	/* wait for character to be ready   */
	;
    return ((*siop->s_read)(siop));	/* read and return it		    */
}


/*
FUNCTION NAME:	s_putc
LEVEL:		2
LIBRARY:	BOUS.LIB
DESCRIPTION:	Sends a byte out the serial port(SIO) pointed to siop.
OTHER MODULES:	None
RETURNS:	void
COMMENTS:	Access to the serial port is through an sio structure.
*/

void s_putc(siop, c)
register SIO *siop;
BYTE c;
{
    while ((*siop->s_xstat)(siop) == 0)	/* wait for UART to be ready	*/
	;
    (*siop->s_send)(siop, c);		/* sent the character and return	*/
}


/*
FUNCTION NAME:	s_inchar
LEVEL:		2
LIBRARY:	BOUS.LIB
DESCRIPTION:	Fetches a byte from the serial port (SIO) pointed to by siop,
		if one is ready; otherwise returns.
OTHER MODULES:	None
RETURNS:	int: the byte if one is ready; otherwise -1.
COMMENTS:	The function can also be written as a macro in STDIO.H.
*/

int s_inchar(siop)
SIO *siop;
{
    if ((*siop->s_rstat)(siop))
	return ((*siop->s_read)(siop));
    return (-1);
}


/*
FUNCTION NAME:	vsetbits
LEVEL:		2
LIBRARY:	BUOS.LIB
DESCRIPTION:	Builds a new bit pattern in the virtual register in v_regp by
		ANDing and ORing, then calls the level-0 _setbits to install
		that bit pattern in the physical UART's register.
RETURNS:	int: 0 if successful; OR_AGE if argument is out of range;
		BAD_FCN if the function is not supported(SIO pointer is NIL);
		BAD_ARG if the requested setting is not supported by hardware
		(mask initialized to -1).
COMMENTS:	More thanone virtual register may be attached to a function
		by the linked list of next pointers.
*/

int vsetbits(siop, rbp, rank)
SIO *siop;
struct vregbits_ *rbp;
BYTE rank;			/* rank of value to install	*/
{
    BYTE tmp;
    if (rbp == NIL)		/* function not supported in hardware	*/
	return (BAD_FCN);
    if (rank >= NUMMASKS)	/* argument range check 		*/
	return (OR_ARG);
    if (rbp->setmask[rank] == -1)   /* argument not supported in hardware   */
	return (BAD_ARG);
    do	{
	siop->v_regp[rbp->vregnum][VIR] &= rbp->resetmask;  /* reset bits   */
	tmp = siop->v_regp[rbp->vregnum][VIR] |= rbp->setmask[rank];
	rbp->now = rank;	    /* update current table rank	*/
	_vsetbits(siop, rbp, tmp);  /* write vreg to the UART		*/
	rbp = rbp->next;	    /* another structure involved?	*/
	} while (rbp != NIL);	    /* no, done 			*/
    return (0);
}


/*
FUNCTION NAME:	vsetbr
LEVEL:		2
LIBRARY:	BUOS.LIB
DESCRIPTION:	Sets the baud rate to the rank element of the structure's
		divisor array. Passes high- and low-order divisors to the
		level-1 function, _vsetbr.
OTHER MODULES:	_vsetbr
RETURNS:	int: 0 if successful; OR_ARG if argument is out of range;
		BAD_FCN if the function is not supported(SIO pointer is NIL);
		BAD_ARG if the requested setting is not supported by hardware
		(divisor is -1).
COMMENTS:	The rank argument is an index for the array of divisor,
		constants for which reside in SIODEF.H. Notice that the low
		byte is passed before the high byte.
*/

int vsetbr(siop, brp, rank)
SIO *siop;
struct vbaud_ *brp;
BYTE rank;			/* rank of value to install		*/
{
    BYTE brlo, brhi;		/* intermediates simplify syntax	*/
    if (brp == NIL)		/* baud rate not supported in hardware	*/
	return (BAD_FCN);
    if (rank >= NUMBAUDS)	/* requested rank out of range		*/
	return (OR_ARG);
    if (brp->divisor[rank][1] == -1 || brp->divisor[rank][0] == -1)
	return (BAD_ARG);	/* requested baud rate not available	*/
    brlo = brp->divisor[rank][1]; /* lo byte of divisor 		*/
    brhi = brp->divisor[rank][0]; /* hi byte of divisor 		*/
    brp->now = rank;		/* update baud in effect		*/
    _vsetbr(siop, brp, brlo, brhi); /* call level-1 to do it		*/
    return (0);
}

/*
FUNCTION NAME:	vset232
LEVEL:		2
LIBRARY:	BUOS.LIB
DESCRIPTION:	Turns selected RS-232 output ON or OFF.
OTHER MODULES:	None
RETURNS:	int: OR_ARG if state argument is other than 0 or 1;
		BAD_FCN if the requested function is not supported in
		hardware; otherwise 0.
COMMENTS:	The test for a NIL pointer  must be the first error test.
		OFF and ON are defined in SIODEF.H
*/

int vset232(siop, p232, state)
SIO *siop;
struct vout232_ *p232;
BOOL state;			/* state in table of mask to use	*/
{
    if (p232 == NIL)
	return (BAD_FCN);
    if ((state != OFF) && (state != ON))    /* rank check		*/
	return (OR_ARG);
    do	{
	siop->v_regp[p232->vregnum][VIR] &= p232->resetmask; /* reset bit*/
	if (state == ON)
	    siop->v_regp[p232->vregnum][VIR] |= ~p232->resetmask; /* set bits*/
	p232->now = state;	/* update current state variable	*/
	_vset232(siop, p232, siop->v_regp[p232->vregnum][VIR]);
	p232 = p232->next;
	} while (p232 != NIL);	/* more structs in linked list? 	*/
    return (0);
}

/*
FUNCTION NAME:	vstat232
LEVEL:		2
LIBRARY:	BUOS.LIB
DESCRIPTION:	Returns status of RS-232 input pointed to by p232.
RETURNS:	int: TRUE if the input is asserted; FALSE if inhibited;
		BAD_FCN if the input is not supported.
COMMENTS:	The call to the level-1 function is through a function
		pointer in the vin232_ structure.
*/

int vstat232(siop, p232)
SIO *siop;
struct vin232_ *p232;		/* pointer to a @[vin232]@ structure	*/
{
    if (p232 == NIL)
	return (BAD_FCN);	/* this input not supported		*/
    p232->now = (*p232->read232)(siop, p232) & p232->mask;
    return (p232->now ? TRUE : FALSE);	/* return status		*/
}

/*
FUNCTION NAME:	s_fputc
LEVEL:		2
LIBRARY:	BUOS.LIB
DESCRIPTION:	Provides "formated" character-by-character serial output
		according to the soctl_ structute.
OTHER MODULES:	eputch
RETURN: 	int: NOECHO if "wait for echo" times out; KEY_CAN if a loop
		(e.g., XON/XOFF) is aborted from keyboard; otherwise 0.
COMMENTS:	Features:
		Raw (no processing)		    High-bit clearing
		Interbyte and interline delay	    Character filter
		XON/XOFF flow control		    Character translation
		End-of-line conversion		    Console output
		Upper/lower-case conversion	    Wait-for-echo
*/

static void s_eputc();

int s_fputc(siop, c)
SIO *siop;
BYTE c;
{
    register struct soctl_ *op; /* saves lots of indexing time		*/
    static int i, txflag;

    txflag = TRUE;
    op = siop->s_octl;
    op->flushflg = FALSE;
    if (op->rawflg)		/* -- RAW: XMIT --			*/
	{
	s_putc(siop, c);
	return (0);
	}
    if (op->asciiflg)		/* -- RESET BIT 7 --			*/
	c = toascii(c);
    if (op->xxmode && s_inchar(siop) == op->xoffc)  /* -- FLOW CONTROL -- */
	for (EVER)		/* paused awaiting release or abort	*/
	    {
	    if (s_keybort(siop))
		return (KEY_CAN);
	    if ((i = s_inchar(siop)) == -1)
		continue;
	    if (op->xxmode == XON_ANY || i == op->xonc)
		break;
	    }
    if (op->filterflg)		/* -- TRANSLATE/REMOVER BYTE -- 	*/
	{
	for (i = op->numfc - 1; i >= 0; i--)
	    if (c == op->filtlist[i][0])
		{
		if (op->filtlist[i][1] == -1)	/* remove		*/
		    txflag = FALSE;
		else
		    c = op->filtlist[i][1];	/* translate		*/
		break;
		}
	}
    if (iscntrl(c) && txflag)	/* -- REMOVE CONTROLS --		*/
	{
	if (op->cntrlflg)
	    {
	    txflag = FALSE;
	    for (i = op->numcc - 1; i >= 0; i--)
		if (c == op->ok_cntrls[i])	/* loop through list	*/
		    {
		    txflag = TRUE;		/* if on list		*/
		    break;
		    }
	    }
	if (txflag)		/* -- END OF LINE TRANSLATION --	*/
	    {
	    switch (op->eolmode)
		{
		case NOXLAT:			/* no conversion	*/
		    break;
		case LF2CR:			/* lf -> cr		*/
		    if (c == LF)
			{
			s_eputc(siop, op->lastc == CR ? LF : CR);
			txflag = FALSE;
			}
		    break;
		case CR2LF:			/* cr -> lf		*/
		    if (c == CR)
			{
			txflag = FALSE;
			op->flushflg = TRUE;
			break;
			}
		    if (op->lastc == CR)
			s_eputc(siop, c == LF ? CR : LF);
		    break;
		case EOL2SPACE:     /* cr and lf converted to spaces	*/
		    if (c == CR || c == LF)	/* crlf = one space	*/
			{
			txflag = FALSE;
			if (c != LF && op->lastc != CR)
			    s_eputc(siop, SP);
			}
		    break;
		case REMOVEEOL:     /* cr and lf not transmitted	*/
		    if (c == CR || c == LF)
			txflag = FALSE;
		    break;
		case LF2CRLF:	    /* lf -> crlf			*/
		    if (c == LF && op->lastc != CR)
			s_eputc(siop, CR);
		    break;
		case CR2CRLF:	    /* cr -> crlf			*/
		    if (c == CR)    /* wait for next byte		*/
			{
			txflag = FALSE;
			op->flushflg = TRUE;
			}
		    if (op->lastc == CR)
			{
			s_eputc(siop, CR);
			if (c != LF)
			    s_eputc(siop, LF);
			}
		    break;
		case CRLF2LF:	    /* crlf -> lf			*/
		    if (c == CR)
			{
			txflag = FALSE;
			op->flushflg = TRUE;
			break;
			}
		    if (op->lastc == CR && c != LF)
			s_eputc(siop, CR);
		    break;
		case CRLF2CR:	    /* crlf -> cr			*/
		    if (op->lastc == CR && c == LF)
			txflag = FALSE;
		    break;
		default:
		    break;
		}		    /* end of end-of-line switch	*/
	    }
	}
    if (txflag)
	{			/* -- FLOW CONTROL --			*/
	switch (op->casemode)	/* -- CONVERT CASE --			*/
	    {
	    case UP_LOW:
		break;
	    case UPPER:
		c = toupper(c);
		break;
	    case LOWER:
		c = tolower(c);
		break;
	    }
	if (op->ib_dlyflg)	/* -- INTERBYTE, -LINE DELAYS --	*/
	    delay(op->ib_dly);
	s_eputc(siop, c);	/* xmit with optional console output	*/
	if (op->wechoflg)	/* -- WAIT FOR REMOTE ECHO --		*/
	    if (s_waitch(siop, op->wecho_dly) == TIMEOUT)
		return (NOECHO);
	if (op->il_dlyflg && c == op->il_dlyc)
	    delay(op->il_dly);	/* end of line delay			*/
	}
    op->lastc = c;		/* save this character			*/
    return(0);
}

/*
FUNCTION NAME:	s_eputc
LEVEL:		2
LIBRARY:	BUOS.LIB
DESCRIPTION:	Outputs the byte to the SIO and, if consflg is set, to
		standard output.
OTHER MODULES:	s_putc
RETURNS:	void
COMMENTS:	Declared static--i.e., local to this file.
*/

static void s_eputc(siop, c)
SIO *siop;
int c;
{
    s_putc(siop, c);
    if (siop->s_octl->consflg)	    /* echo to standard out		*/
	putchar(c);
}

/*
FUNCTION NAME:	s_fgetc
LEVEL:		2
LIBRARY:	BUOS.LIB
DESCRIPTION:	Provides "formatted" character-by-character serial input
		according to the sictl_ structure pointed to by the SIO.
OTHER MODULES:	None
RETURNS:	int: TIMEOUT if t_outflg is TRUE and no byte arrives in
		t_out Tsecs; IGNORE if an EOL decision is pending;
		otherwise the byte received.
COMMENTS:	Functions supported:
		Switchable timeout mode     Raw mode
		Control character filter    Character translation/filter
		End-of-line conversion	    Console output
		XON/XOFF flow control	    Upper/lower-case conversion
		Echo-back to sender	    Hifh bit clearing
*/

int s_fgetc(siop)
SIO *siop;
{
    register struct sictl_ *ip;     /* registers and statics for speed	*/
    static int retbyte, i, c;
    static BOOL rxflag;
    if (siop->s_ictl->rawflg)	    /* do raw processing asap		*/
	return (s_inchar(siop));
    ip = siop->s_ictl;		    /* save lots of indexing time	*/

    if (ip->rbakflg)		    /* if byte pending			*/
	{
	ip->rbakflg = FALSE;
	if (ip->consflg)
	    putchar(ip->rbakc);
	return (ip->rbakc);
	}
    c = (ip->t_outflg) ? s_waitch(siop, ip->t_out) : s_inchar(siop);
    if (c == TIMEOUT)
	return (TIMEOUT);
    retbyte = -1;		    /* postpone initialization		*/
    rxflag   = TRUE;
    if (ip->ekobakflg)		    /* -- ECHO TO SENDER --		*/
	s_putc(siop, c);
    if (ip->asciiflg)		    /* -- RESET BIT 7 --		*/
	c = toascii(c);
    if (ip->filterflg)		    /* -- TRANSLATE/REMOVE BYTE --	*/
	{
	for (i = ip->numfc - 1; i >= 0; i--)
	    if (c == ip->filtlist[i][0])
		{
		if (ip->filtlist[i][1] == -1)	/* remove		*/
		    rxflag = FALSE;
		else
		    c = ip->filtlist[i][1];	/* translate		*/
		break;
		}
	}
    if (iscntrl(c) && rxflag)	    /* -- REMOVE CONTROLS --		*/
	{
	if (ip->cntrlflg)
	    {
	    rxflag = FALSE;
	    for (i = ip->numcc-1; i >= 0; i--)
		if (c == ip->ok_cntrls[i])	/* loop through list	*/
		    {
		    rxflag = TRUE;		/* if on list		*/
		    break;
		    }
	    }
	}
    if (rxflag) 		    /* -- END-OF-LINE TRANSLATION --	*/
	{
	switch (ip->eolmode)
	    {
	    case NOXLAT:	    /* no conversion			*/
		break;
	    case LF2CR: 	    /* lf->cr				*/
		if (c == LF)
		    retbyte = CR;
		break;
	    case CR2LF: 	    /* cr->lf				*/
		if (c == CR)
		    retbyte = LF;
		break;
	    case EOL2SPACE:	    /* cr & lf converted to spaces	*/
		if (c == CR || c == LF)     /* crlf = one space 	*/
		    if (c != LF && ip->lastc != CR)
			rxflag = FALSE;
		    else
			retbyte = SP;
		break;
	    case REMOVEEOL:	    /* cr and lf not transmitted	*/
		if (c == CR || c == LF)
		    rxflag = FALSE;
		break;
	    case CR2CRLF:	    /* cr->crlf 			*/
		if (c == CR)
		    {
		    ip->rbakflg = TRUE;
		    ip->rbakc	= LF;
		    }
		break;
	    case CRLF2LF:	    /* crlf->lf 			*/
		if (c == CR && ip->lastc != CR)
		    rxflag = FALSE;
		else
		    retbyte = (ip->lastc == CR) ? LF : CR;
		break;
	    case CRLF2CR:	    /* crlf->cr 			*/
		if (c == CR && ip->lastc != CR)
		    rxflag = FALSE;
		else
		    retbyte = CR;
		break;
	    default:
		break;
	    }			    /* end of end-of-line switch	*/
	}
    if (rxflag) 		    /* -- CASE CONVERSION --		*/
	switch (ip->casemode)
	    {
	    case UP_LOW:
		break;
	    case UPPER:
		c = toupper(c);
		break;
	    case LOWER:
		c = tolower(c);
		break;
	    }
    ip->lastc = c;		    /* save this character		*/
    if (retbyte != -1)	    /* if byte to return is not byte received	*/
	c = retbyte;
    if (ip->consflg && rxflag)
	putchar(c);
    return (rxflag ? c : IGNORE);
}

/*
FUNCTION NAME:	s_iflush
LEVEL:		2
LIBRARY:	BUOS.LIB
DESCRIPTION:	Returns pending byte from the s_octl's "look-back" buffer.
OTHER MODULES:	None
RETURNS:	int: byte if available; otherwise -1.
*/

int s_iflush(siop)
SIO *siop;
{
    return ((siop->s_ictl->rbakflg) ? siop->s_ictl->rbakc : -1);
}
