
/***************************************************************************

    library for Hayes Smartmodem and compatible modems.

    This library contains follow functions and arrays:

    LEVEL 1:
	m_cmd		Builds a smart-modem command in the modem structure's
			command buffer, then outputs it to the SIO.
	m_quiz		Fetches bytes from the modem. The maximum number of
			bytes to be fitched is specified in numc.
	m_result	A two-by-three array of string poinyers containing
			all valid smart-modem responses.
	m_isreply	Converts result codes to an integer code.

    LEVEL 2:
	m_query 	Outputs a command, then fetches the modem's response.
			The maximum number of bytes to be fetched is
			specified in numc.
	m_qcmd		Outputs the modem command, then waits for a response
			of either OK or 0.
	m_getint	Query modem for a positive integer response.
	m_wait4dcd	Waits for modem response. The wait loop can be
			terminated from the keyboard.
			Response from the modem is converted to an integer
			code and returned. Codes are in MODEM.H.
	m_gocmd 	Forces the modem from online to command state.

    LEVEL 3:
	m_reset 	Attempts to reset the modem to default condition.
	m_whoru 	Queries the modem for its product and ROM codes and
			stores them in smid and romck, respectively, in the
			modem structure. Also installs a modem
			indentification code in modtype for later use.
	m_config	Configures the modem as described in the modem
			structure.
	m_dial		Dials the number specified and waits for the return
			of a result code.
	m_redial	Redials a previously dialed number.
	m_answer	Answers an incoming call and optionally adhusts the
			baud rate to the rate of the call.
	m_hup		Hangs up the phone.
	m_warndcd	Issues a warning and overrides the prompt if the
			current DCD state does not match the argument.
	m_codes 	Strings corresponding to smart-modem response codes.
	m_names 	Strings corresponding to modem id codes.

***************************************************************************/

#include    "ascii.h"
#include    "stdio.h"
#include    "siodef.h"
#include    "sioctl.h"
#include    "modem.h"

/*
FUNCTION NAME:	m_cmd
LEVEL:		1
LIBRARY:	MODEM.LIB
DESCRIPTION:	Bulids a smart-modem command in the modem structure's
		command buffer, then outputs it to the SIO.
OTHER MODEULES: s_ocntrl, s_opush, s_opop
RETURNS:	void
COMMENTS:	The f0llowing output characteristics are forced: a fixed
		interbyte delay and upper-case output. Commands are echoed
		to the console based upon the cmnechoflg structure member.
*/

void m_cmd(siop, cmdstr)
SIO *siop;
BYTE *cmdstr;
{
    struct modem *smp = siop->sm;	/* for notational convenience	*/
    BOOL dtr_stat;			/* save state of dtr		*/
    if ((  dtr_stat = dtrnow(siop)) == OFF) /* if dtr inhibited now ... */
	setdtr(siop, ON);		/* assert it			*/
    s_opush(siop);			/* preserve output structure	*/
    s_ocntrl(siop, SET, CASEMODE, UPPER);   /* convert to upper case	*/
    s_ocntrl(siop, SET, ASCIIFLAG, ON);     /* strip high bits		*/
    s_ocntrl(siop, SET, IB_DLYFLAG, ON);    /* interbyte delay		*/
    s_ocntrl(siop, SET, IB_DLY, smp->ibdelay);
    s_ocntrl(siop, SET, CONSFLAG, smp->cmdechoflg); /* echo commands?	*/
    strcpy(smp->cmdbuff, smp->csi);	    /* start with CSI prefix	*/
    strcat(smp->cmdbuff, smp->quietcmd);    /* append result code mode	*/
    strcat(smp->cmdbuff, cmdstr);	    /* append commend to buffer */
    strcat(smp->cmdbuff, smp->eocstr);	/* append terminator to buffer	*/
    //printf("Dialstr = %s\n", smp->cmdbuff);	/* for debug			*/
    s_fputs(siop, smp->cmdbuff);	/* output buffer to SIO 	*/
    if (dtr_stat == OFF)		/* restore DTR			*/
	setdtr(siop, OFF);
    s_opop(siop);			/* restore output structure	*/
}

/*
FUNCTION NAME:	m_quiz
LEVEL:		1
LIBRARY:	MODEM.LIB
DESCRIPTION:	Fetches bytes form the modem. The maximem number of bytes to
		be fetched is specified in numc.
OTHER MODULES:	s_ipush, s_ipop, s_icntrl, s_fgets
RETURNS:	char *: pointer to the modem's response; NIL if no response.
COMMENTS:	The timeout interval os derived from the modem structure.
*/

char *m_quiz(siop, numc)
SIO *siop;
short numc;			    /* # of bytes wanted in response	*/
{
    static char quizbuff[QBUFSIZ];  /* place modem's replay here        */
    char *quizp = quizbuff;	    /* pointer value to return		*/
    if (numc >= QBUFSIZ)	    /* protect against buffer overrun	*/
	numc = QBUFSIZ - 1;
    s_ipush(siop);
    s_icntrl(siop, SET, T_OUTFLAG, ON);
    s_icntrl(siop, SET, T_OUT, siop->sm->cmdt_out);
    if (s_fgets(siop, quizbuff, numc) == 0) /* get reply into buffer	*/
	quizp = NIL;		    /* return a NIL if 0 bytes returned */
    s_ipop(siop);
    return (quizp);
}

/*
ARRAY NAME:	m_result
TYPE:		char *
LEVEL:		3
LIVRARY:	MODEM.LIB
DESCRIPTION:	A two-by-three array of string pointers containing all valid
		smart-modem respoises.
COMMENTS:	The order of serach on this array is critical.
*/

char *m_result[][3] =
{
    { "\r\nOK\r\n",		"\r0\r",    "0\r"   },	/* must be first    */
    { "\r\nCONNECT\r\n",	"\r1\r",    "1\r"   },
    { "\r\nRING\r\n",		"\r2\r",    "2\r"   },
    { "\r\nNO CARRIER\r\n",	"\r3\r",    "3\r"   },
    { "\r\nERROR\r\n",		"\r4\r",    "4\r"   },
    { "\r\nCONNECT 1200\r\n",	"\r5\r",    "5\r"   },
    { "\r\nNO DIALTONE\r\n",	"\r6\r",    "6\r"   },	/* Hayes	*/
    { "\r\nNO DIAL TONE\r\n",	"\r6\r",    "6\r"   },	/* variant	*/
    { "\r\nBUSY\r\n",		"\r7\r",    "7\r"   },
    { "\r\nNO ANSWER\r\n",	"\r8\r",    "8\r"   },
    { NIL,			NIL,	    NIL     },
    { "\r\nCONNECT 2400\r\n",	"\r10\r",   "10\r"  },
};
int numrows = sizeof(m_result) / (3 * sizeof(m_result[0][0]));	/* cardinal */

/*
FUNCTION NAME:	m_isreply
LEVEL:		1
LIBRARY:	MODEM.LIB
DESCRIPTION:	Converts result codes to an integer code.
OTHER MODULES:	substr, given in YOUR.LIB
RETURNS:	int: an integer result code or -1 if the code is nor valid.
COMMENTS:	This function works correctly with either long or single-
		digit result codes. The word codes are searched first, then
		the digit codes. To avoid an erroneous match with any message
		ending in '0', the array is searched backward.
*/

int m_isreply(str)
char *str;
{
    extern char *m_result[][3];
    extern int numrows;
    int row;		    /* init in for loop to start at last row	*/
    int col = 0;
    do	{
	for (row = numrows - 1; row >= 0; row--)    /* start with last row  */
	    if (substr(str, m_result[row][col]))
		return (atoi(m_result[row][2]));
	} while (++col < 3);
    return (-1);			/* -1 if not a substring	*/
}

/*
FUNCTION NAME:	m_query
LEVEL:		2
LIBRARY:	MODEM.LIB
DESCRIPTION:	Outputs a command, theb fetches the modem's response. The
		maximum number of bytes to be fetched is specified in numc.
OTHER MODULES:	s_clrsio, m_cmd, m_quiz.
RETURNS:	char *: pointer to the modem's response; NIL if no response.
COMMENTS:	Any echoed bytes are cleared from the SIO between issuing
		the command and fetching the reply.
*/

char *m_query(siop, querystr, numc)
SIO *siop;
char *querystr;
short numc;				/* # bytes wanted in response	*/
{
    char *m_quiz();
    if (s_clrsio(siop, CMDBUFSIZ) == NUL)   /* clear SIO		*/
	return (NIL);
    m_cmd(siop, querystr);		/* send desired command 	*/
    s_clrsio(siop, strlen(querystr));	/* clesr echo			*/
    return m_quiz(siop, numc);		/* get and return pointer to reply  */
}

/*
FUNCTION NAME:	m_qcmd
LEVEL:		2
LIBRARY:	MODEM.LIB
DESCRIPTION:	Outputs the modem command, then waits for a response of
		either OK or 0.
OTHER MODULES:	m_query, m_isreply
RETURNS:	int: M_SUCCEED if command is auccessful; otherwise M_FAIL.
COMMENTS:	Not intendedto fetch result codes from dialing operations!
*/

int m_qcmd(siop, cmdstr)
SIO *siop;
BYTE *cmdstr;
{
    int retries;
    char *reply, *m_query();
    retries = siop->sm->cmdretry;	/* from structure		*/
    do	{				/* do at least once		*/
	if ((reply = m_query(siop, cmdstr, MAXREPLY)) != NIL)
	    if (m_isreply(reply) == M_OK)
		return (M_SUCCEED);
	} while (--retries > 0);
    return (M_FAIL);
}

/*
FUNCTION NAME:	m_getint
LEVEL:		2
LIBRARY:	MODEM.LIB
DESCRIPTION:	Query modem for a positive integer response.
OTHER MODULES:	m_query, isdigit
RETURNS:	int: the integer or -1 if none found.
COMMENTS:	Will not convert an integer less than two digits.
*/

int m_getint(siop, cmdstr)
SIO *siop;
char *cmdstr;
{
    char *buffp, *m_query();
    if ((buffp = m_query(siop, cmdstr, 10)) != 0)
	for (; *buffp != NUL; buffp++)
	    if (isdigit(*buffp) && isdigit(*(buffp+1)))
		return (atoi(buffp));
    return (-1);
}

/*
FUNCTUIN NAME:	m_wait4dcd
LEVEL:		2
LIBRARY:	MODEM.LIB
DESCRIOPTION:	Waits for modem response. The Wait loop can be terminated
		from the keyboard. Response from the modem is conveted to
		an integer code and returned. Codes are in MODEM.H.
OTHER MODULES:	m_isreply, s_clrsio, s_fgets, s_icntrl, s_ipop, s_ipush,
		s_keybort
RETURNS:	int: the modem's integer response code to dialing(M_NODCD) if
		no response occurs within the specified time; M_USRCAN if
		user presses abort key; M_SIOERR if SIO will not clear;
		or M_BADMSG if the response is invalid.
COMMENTS:	The duration of the wait for carrier is determined by the
		Smartmodem's S-register 7, the value of which is kept in the
		modem structure member dcdwait.
*/

int m_wait4dcd(siop)
SIO *siop;
{
    int retcode = -1;		    /** must initialize this !		*/
    char waitbuff[MAXREPLY + 1];
    if (s_clrsio(siop, CMDBUFSIZ) == NUL)   /* clear SIO		*/
	return (M_SIOERR);
    s_ipush(siop);
    s_icntrl(siop, SET, T_OUTFLAG, ON);     /* turn on timout mode	*/
    s_icntrl(siop, SET, T_OUT, _0_SEC_2);   /* timeout for s_fgets	*/
    do	{			    /* wait for response from modem	*/
	if (s_fgets(siop, waitbuff, MAXREPLY) != 0)
	    if ((retcode = m_isreply(waitbuff)) == -1)
		retcode = M_BADMSG;	/* -1 if not a result code	*/
	if (s_keybort(siop))
	    {
	    retcode =  M_USRCAN;	/* canceled call from keyboard	*/
	    break;
	    }
	} while (retcode == -1);
    s_ipop (siop);
    return (retcode);
}

/*
FUNCTION NAME:	m_gocmd
LEVEL:		2
LIBRARY:	MODEM.LIB
DESCRIPTION:	Forces the modem from online to command state.
OTHER MODEULES: s_ipush, delay, s_icntrl, s_ipop
COMMENTS:	Three "escape" characters are sent, surrounded by the quiet
		interval given by the modem structure member guardtime.
		The smart-modem escape character is kept in S-register 2 and
		defaults to '+' (ASCII 28).
*/

int m_gocmd(siop, escchar)
SIO *siop;
char escchar;
{
    int i;
    int errcode = M_FAIL;
    char escbuff[MAXREPLY];
    s_ipush(siop);
    s_icntrl(siop, SET, T_OUTFLAG, ON);     /* force timeout on input	*/
    s_icntrl(siop, SET, T_OUT, siop->sm->guardtime +1); /* force timeout*/
    delay(siop->sm->guardtime);
    for (i = 3; i > 0; --i)
	s_fputc(siop, escchar);
    if (s_fgets(siop, escbuff, 10) != 0)    /* get reply into buffer	*/
	errcode = m_isreply(escbuff) == M_OK;	/* TRUE/FALSE		*/
    s_ipop(siop);
    return (errcode);
}

/*
FUNCTION NAME:	m_reset
LEVEL:		3
LIBRARY:	MODEM.LIB
DESCRIPTION:	Attempts to reset the modem to default condition.
OTHER MODULES:	delay, setgpo1, setdtr, setbaud, baudnow, m_gocmd, m_qcmd,
		m_cmd
RETURNS:	int: M_SUCCESS if reset is successful; M_FAIL if user aborts.
COMMENTS:	Hardware reset is used if user installs modem as a "B"(board)
		type. If user elects to terminate the reset, rs232flg in the
		modem structure is made FALSE.
*/

int m_reset(siop)
SIO *siop;
{
    extern char *brstr[];
    struct modem *smp = siop->sm;
    int i, j, errcode, user_redo;
    int tmptries, tmptime, tmpbr;	    /* temporary storage	*/
    if (s_clrsio(siop, CMDBUFSIZ) == NUL)   /* slear SIO of garbage	*/
	return (M_SIOERR);
    tmptries = smp->cmdretry;	    /* save number of retries		*/
    tmptime = smp->cmdt_out;	    /* save command response time out	*/
    tmpbr = baudnow(siop);	    /* save current baud rate		*/
    setdtr(siop, OFF);		    /* inhibit DTR to force comatose mode   */
    delay(_0_SEC_3);		    /* a bit of DTR delay		*/
    setdtr(siop, ON);		    /* assert DTR to enable command mode    */
    smp->cmdretry = 2;		    /* make m_qcmd retry only twice	*/
    smp->cmdt_out = _1_SEC_0 * 2;   /* allow at least 2 secs for reset	*/
    user_redo = FALSE;		    /* TRUE = try online escape ('+++') */
    for (EVER)
	{
	for (i = 0; smp->baudtry[i] != -1; ++i) /* at every baud rate listed	*/
	    {
	    setbaud(siop, smp->baudtry[i]);
	    printf("Trying Reset at %s\n", brstr[smp->baudtry[i]]);
	    for (j = tmptries; j > 0; --j)	/* issue reset command	*/
		{
		smp->rs232flg = TRUE;		/* may be modified below    */
		if (user_redo && j <= 2)	/* 2nd time through; modem...	*/
		    {		/* ...maybe online with DTR is disabled     */
		    m_gocmd(siop, '+'); 	/* default escape character */
		    m_cmd(siop, "H0");		/* issue hangup command     */
		    m_gocmd(siop, atoi(&smp->escape[3]));   /* ditto local  */
		    m_cmd(siop, "H0");		/* hangup command again     */
		    smp->rs232flg = FALSE;	/* DTR doesn't work         */
		    }
		if ((errcode = m_qcmd(siop, "Z")) == M_SUCCEED)
		    {
		    smp->baudfound = smp->baudtry[i];	/* save rate	*/
		    goto EXIT;
		    }
		if (smp->iscard == TRUE)    /* if installed as a B-type */
		    {
		    setgpo1(siop, ON);	    /* GPO1 resets B-type modem */
		    delay(_0_SEC_05);	    /* leave on a while 	*/
		    setgpo1(siop, OFF);     /* release reset		*/
		    }
		}
	    }
	puts("\007\nModem won't reset. Press any key to retry, ESC to quit");
	if (getch() == ESC)
	    {
	    errcode = M_USRCAN;
	    smp->rs232flg = FALSE;	/* broadcast to other function	*/
	    break;
	    }
	if (user_redo == FALSE) /* next time through, try online escape */
	    user_redo = !user_redo;
	}

    EXIT:
    setdtr(siop, OFF);
    smp->cmdt_out = tmptime;		/* restore from temporaries	*/
    smp->cmdretry = tmptries;
    setbaud(siop, tmpbr);
    return (errcode);
}

/*
FUnCTION NAME:	m_whoru
LEVEL:		3
LIBRARY:	MODEM.LIB
DESCRIPTION:	Queries the modem for its product and ROM codes and stores
		them in smid and romck, respectively, in the modem structure.
		Also installs a modem identification code in modtype for
		later use.
OTHER MODULES:	setdtr, m_getint, m_query, s_ipush, s_ipop, s_ictrl
RETURNS:	int: M_FAIL if modem does not respond to the query or if the
		response is unidentifiable; otherwise M_SUCCEED
COMMENTS:	REMOVEEOL is forced to guarantee the correct number of bytes.
		Only the first two digits in the product ID are used.
*/

int m_whoru(siop)
SIO *siop;
{
    struct modem *smp = siop->sm;
    int errcode = M_SUCCEED;
    int id;
    setdtr(siop, ON);
    if ((id = smp->smid = m_getint(siop, "I0")) == -1)	/* error	*/
	errcode = M_FAIL;		/* the only possible failure	*/
    if ((smp->romck = m_getint(siop, "I1")) == -1)
	errcode = M_FAIL;
    if (id >= 100)	    /* reduce id to two digits for switch below */
	id /= 10;
    if (smp->modtype == VIRGIN)     /* modtype may already be installed */
	switch (id)
	    {
	    case 12:
		smp->modtype = ID_12;
		break;
	    case 13:
		smp->modtype = ID_12PLUS;
		break;
	    case 24:
		smp->modtype = ID_24;
		break;
	    default:
		smp->modtype = ID_300;
	    }
    setdtr(siop, OFF);
    return (errcode);
}

/*
FUNCTION NAME:	m_config
LEVEL:		3
LIBRARY:	MODEM.LIB
DESCRIPTION:	Configures the modem as described in the modem structure.
OTHER MODULES:	srtbaud
RETURNS:	int: M_SUCCEED or M_FAIL
COMMENTS:	Builds two configuration strings: pri_config containing
		commands that are recognized by all Smartmodems, and
		sec_config containing commands that are supported by only
		some. Note that sec_config must be sent first.
*/

int m_config(siop)
SIO *siop;
{
    struct modem *smp = siop->sm;
    int retval;
    setdtr(siop, ON);
    strcpy(smp->pri_config, "");    /* null out primary config buffer	*/
    strcpy(smp->sec_config, "");    /* null out secondary config buffer */
    switch (smp->modtype)	    /* build sec_config based on ID code*/
	{
	case ID_300:
	    smp->baudmax = BAUD300;		/* maximum baud rate	*/
	    break;
	case ID_12:
	    smp->baudmax = BAUD1200;
	    smp->xcmd[1] = '1'; 		/* only X1 supported	*/
	    strcat(smp->sec_config, smp->xcmd);
	    break;
	case ID_12PLUS:
	    smp->baudmax = BAUD1200;
	    strcat(smp->sec_config, smp->xcmd);
	    strcat(smp->okalpha, "/!@W");	/* new dial modifiers	*/
	    strcat(smp->sec_config, smp->speakvol);
	    break;
	case ID_24:
	    smp->baudmax = BAUD2400;
	    strcat(smp->sec_config, "&F");	/* recall factory config    */
	    strcat(smp->sec_config, smp->firmdtr);  /* dtr mode 	*/
	    strcat(smp->sec_config, smp->firmdcd);  /* dcd mode 	*/
	    strcat(smp->sec_config, smp->xcmd); /* extended commands	*/
	    strcat(smp->okalpha, "/!@W");	/* now dial modifiers	*/
	    strcat(smp->sec_config, smp->speakvol);
	    break;
	}
    m_qcmd(siop, smp->sec_config);	    /* secondary must be first! */
    strcat(smp->pri_config, smp->dialmode); /* pulse or tone dialing	*/
    strcat(smp->pri_config, smp->echocmd);  /* modem-echo in command mode   */
    strcat(smp->pri_config, smp->verbosecmd);	/* long or short result codes	*/
    strcat(smp->pri_config, smp->autoans);  /* # rings on auto answer	*/
    strcat(smp->pri_config, smp->escape);   /* escape character 	*/
    strcat(smp->pri_config, smp->dcdwait);  /* secs to wait for carrier */
    strcat(smp->pri_config, smp->dtwait);   /* secs to wait for dialtone    */
    strcat(smp->pri_config, smp->speaker);  /* speaker on/off		*/
    retval = m_qcmd(siop, smp->pri_config); /* send primary to modem	*/
    switch (smp->baudmode)		    /* set baud rate		*/
	{
	case BAUD_DEF:			/* no change			*/
	    break;
	case BAUD_MAX:			/* to modem's highest           */
	    setbaud(siop, smp->baudmax);
	    m_qcmd(siop, "");		/* tell baud rate to modem	*/
	    break;
	case BAUD_FOUND:		/* rate where modem responded	*/
	    if (smp->modtype == ID_300 && smp->baudfound > BAUD300)
		setbaud(siop, BAUD300); /* special case 		*/
	    else
		setbaud(siop, smp->baudfound);	/* where reset found it */
	    m_qcmd(siop, "");		/* tell baud rate to modem	*/
	    break;
	}
    if (retval == M_SUCCEED)
	smp->configokflg = TRUE;	/* broadcast init		*/
    setdtr(siop, OFF);			/* inhibit DTR			*/
    return (retval);			/* return 0 if "OK"		*/
}

/*
FUNCTION NAME:	m_dial
LEVEL:		3
LIBRARY:	MODEM.LIB
DESCRIPTION:	Dials the number spedified and waits for the retrun of a
		result code.
OTHER MODEULES: isdcd, m_cmd, s_ipush, s_icntrl, s_putc, strcpy, strcat,
		strlen, setdtr, m_wait4dcd
RETURNS:	int: result code returned from dial operation.
COMMENTS:	The duration of the wait for the response is given in the
		dcdwait structure member. The filtered dial string is built
		into the modem structure for use by redial.
*/

int m_dial(siop, phonenum)
SIO *siop;
char phonenum[];
{
    struct modem *smp = siop->sm;
    int resultcode;
    char dialstr[CMDBUFSIZ];	    /* build dial command here		*/
    char *dbp = smp->dialbuff;
    do	{			    /* filter string for illegal bytes	*/
	if (isdigit(*phonenum) || index(smp->okalpha, *phonenum))
	    *dbp++ = *phonenum;     /* copy valid characters to dialbuff    */
	} while(*++phonenum != NUL);
    strcpy(dialstr, smp->dialcmd);  /* build dial string: dial command..*/
    strcat(dialstr, smp->dialmode); /* dial by tone or pulse		*/
    strcat(dialstr, smp->dialbuff); /* ... then the number itself then..*/
    strcat(dialstr, smp->enddial);  /* ... spedial dial commands e.g. ';'   */
    if (strlen(smp->dialbuff) == 0) /* exit if string is NIL		*/
	return (M_NILNUM);
    setdtr(siop, ON);
    m_cmd(siop, dialstr);	    /* output command to SIO		*/
    resultcode = m_wait4dcd(siop);  /* go wait for carrier		*/
    switch (resultcode)
	{
	case M_NODCD:		    /* no carrier (or timeout)		*/
	case M_NOTONE:		    /* no dial tone heard		*/
	case M_BUSY:		    /* busy signal heard		*/
	case M_NOANSW:		    /* silence not heard after @	*/
	    setdtr(siop, OFF);	    /* return to command state		*/
	    break;
	case M_USRCAN:		    /* call cancelled from keybuard	*/
	    setdtr(siop, OFF);	    /* return to command state		*/
	    s_putc(siop, SP);	    /* cancel call			*/
	    break;
	}
    return (resultcode);	    /* exit with DTR still high 	*/
}

/*
FUNCTION NAME:	m_redial
LEVEL:		3
LIBRARY:	MODEM.LIB
DESCRIPTION:	Redials a previously dialed number.
OTHER MODULES:	m_dial
RETURNS:	int: response code as described in MODEM.H.
COMMENTS:	The previously dialed number is left in the modem structure
		by m_dial.
*/

int m_redial(siop)
SIO *siop;
{
    return (m_dial(siop, siop->sm->dialbuff));
}

/*
FUNCTION NAME:	m_answer
LEVEL:		3
LIBRARY:	MODEM.LIB
DESCRIPTION:	Answers an incoming call and optionally adjusts the baud
		rate to the rate of the call.
OTHER MODULES:	setdtr, m_cmd, m_wait4dcd, setbaud, s_clrsio, m_quiz,
		m_isreply, delay.
RETURNS:	int: integer answer code from m_wait4dcd.
COMMENTS:	The duration of the wait for the carrier is given in the
		dcdwait structure member. Waits for a RING response from the
		modem before answering.
*/

int m_answer(siop)
SIO *siop;
{
    char *ringmsg, *m_quiz();
    int anscode, newbaud, tmptime;
    if (s_clrsio(siop, CMDBUFSIZ) == NUL)   /* clear SIO of garbage	*/
	return (M_SIOERR);
    tmptime = siop->sm->cmdt_out;	/* save command response time out   */
    siop->sm->cmdt_out = _1_SEC_0 * 10; /* long time between rings	*/
    setdtr(siop, ON);
    ringmsg = m_quiz(siop, 9);
    siop->sm->cmdt_out = tmptime;	/* restore command time out	*/
    if (ringmsg == NIL) 		/* np response			*/
	return (M_NORING);
    if (m_isreply(ringmsg) != M_RING)	/* wrong response		*
	return (M_NORING);
    delay(_0SEC_3);		    /* wait for smartmodem to get ready	*/
    m_cmd(siop, siop->sm->anscmd);  /* ok to issue answer command	*/
    anscode = m_wait4dcd(siop);
    if (siop->sm->autobrflg == ON)  /* change baud to incoming call	*/
	{
	switch (anscode)
	    {
	    case M_CONN3:
		newbaud = BAUD300;
		break;
	    case M_CONN12:
		newbaud = BAUD1200;
		break;
	    case M_CONN24:
		newbaud = BAUD2400;
		break;
	    }
	if (baudnow(siop) != newbaud)	/* baud may already be right	*/
	    setbaud(siop, newbaud);
	}
    switch (anscode)
	{
	case M_USRCAN:
	    s_putc(siop, SPACE);
	case M_NODCD:
	case M_SIOERR:
	    setdtr(siop, OFF);
	}
    return (anscode);
}

/*
FUNCTION NAME:	m_hup
LEVEL:		3
LIBRARY:	MODEM.LIB
DESCRIPTION:	Hangs up the phone.
OTHER MODULES:	delay, setdtr, isdcd, m_gocmd, m_qcmd
RETUTNS:	int: M_SUCCEED or M_FAIL
COMMENTS:	if inhibiting DTR does not force hang-up, an online escape
		command is issued to force the modem to commend mode where
		an explicit hang-up command can be issued.
*/

int m_hup(siop)
SIO *siop;
{
    int err;
    setdtr(siop, OFF);
    delay(_0_SEC_3);		    /* allow for a little delay 	*/
    if ((err = isdcd(siop)))	    /* if DTR is still up		*/
	{
	m_gocmd(siop, atoi(&siop->sm->escape[3]));  /* escape to cmd state  */
	err = m_qcmd(siop, siop->sm->hupcmd);	    /* and hang up	*/
	}
    return (err);
}

/*
FUNCTION NAME:	m_warndcd
LEVEL:		3
LIBRARY:	MODEM.LIB
DESCRIPTION:	Issues a warning and overrides the prompt if the current DCD
		state does not match the argument.
OTHER MODULES:	isdcd
RETURNS:	int: M_SUCCEED if the DCD state is OK, if the user elects to
		ignore the DCD state, or if the rs232flag in the modem
		structure is TRUE; otherwise M_FAIL.
COMMEMTS:	The rs232flg may be FALSE at the user's discretion or
		automatically if any phase of modem reset fails.
*/

int m_warndcd(siop, warnmode)
SIO *siop;
int warnmode;
{
    int c = -1;
    char *warnp;
    if (siop->sm == NIL || siop->sm->rs232flg == FALSE)
	return (M_SUCCEED);	/* no modem structure or DCD not supported  */
    warnp = (warnmode) ? "no" : "a";
    if (isdcd(siop) != warnmode)
	{
	printf("\nTests indicate %s carrier present.\
		    OK to ignore? \007 (Y/N):", warnp);
	while (c != 'N' && c != 'Y')
	    if ((c = toupper(getch())) == 'N')
		return (M_FAIL);
	}
    return (M_SUCCEED);
}

/*
ARRAY NAME:	m_codes
LEVEL:		3
LIBRARY:	MODEM.LIB
DESCRIPTION:	Strings corresponding to smart-modem response codes.
COMMENTS:	The smart-modem response codes--e.g., M_CONN12--are defined
		in MODEM.H.
*/

char *m_codes[] =
{
    "Successful",
    "300 baud connection established",
    "Call incoming",
    "no connection established",
    "Command error",
    "1200 baud connextion established",
    "No dialtone",
    "Dialed number is busy",
    "No answer",
    "?????",			    /* reserved */
    "2400 baud connection established",
    "No ring detected",
    "Unknown smartmodem response",
    "Call already in progress",
    "Blank dial string",
    "Operation cancelled by user",
    "Cannot clear SIO"
};

/*
ARRAY NAME:	m_names
LEVEL:		3
LIBRARY:	MODEM.LIB
DESCRIPTION:	Strings corresponding to modem id codes.
COMMENTS:	The modem id codes--e.g., ID_1200--are defined in MODEM.H.
*/

char *m_names[] =		/* name for each id code		*/
{
    "Unknown type",
    "300",
    "1200/1200B",
    "1200+/1200B+",
    "2400/2400B",
    "Generic 1200 clone",
    "Generic 2300 clone",
};
