
/****************************************************************************

    Level 3 library for serial I/O.

    This library contains follow functions and arrays

	s_puts		transmitting a string
	s_clrsio	Clears numbytes bytes from UART's receive buffers.
	setparity
	setdatalen
	setstops
	paritynow
	dlnow
	stopsnow
	setbaud
	baudnow

	sbstr		string to support stop bits (data format)
	parstr		string to support parity    (	  "     )
	dlstr		string to support data length(	  "      )
	brstr		string to support baud rate (	  "     )
	errstr		string to support error messages

	setrts		These functions set or reset their respective RS-232
	setdtr		output pins.
	setgpo1
	setgpo2
	setbk

	dtrnow		These functions fetch current state of RS-232
	rtsnow		outputs.
	gpo1now
	gpo2now

	iscts		These functions fetch the state of an RS-232 input.
	isdsr
	isdcd
	isri
	isgpi1
	isgpi2

	xmiton		These functions enable transmitter and receiver;
	rcvon		enable transmitter handshake on CTS and receiver
	txhson		handshake on DCD; set clock factor.
	rxhson
	setclk

	s_open		Prepares a serial device for I/O.
	s_close 	Restores the SIO pointed to by siop to its original
			condition.
	s_keybort	Checks the keyboard to see if the designated abort
			key has been typed.

	s_fputs 	Outputs a string to the SIO under control of the
			soctl_ structure.

	s_puts		Transmits a string from the SIO without control
			through the soctl_ structure.

	//s_printf	"Printf-style" output through the SIO.

	s_ocntrl	Sets or fetched the value of the flag, delay, and
			mode members of an sictl_ input control structure.

	s_opush 	Save and restore s_octl structures on local stack.
	s_opop

	s_icntrl	Sets or fetches the value of the flag, delay, and
			mode members of an sictl_ input control structure.

	s_ipush 	Save and restore s_ictl structure on the local stack.
	s_ipop

	s_fgets 	Inputs a NUL- terminated string of numc bytes into
			buffer.

	s_fgetln	Inputs a NUL-terminated string of bytes into buff.
			Input is terminated when the specified delimiter is
			encountered.
****************************************************************************/

#include "stdio.h"
#include "siodef.h"
#include "sioctl.h"

/*
FUNCTION NAME:	s_puts
LEVEL:		3
LIBRARY:	SIO.LIB
DESCRIPTION:	Writes a NUL-terminated string to the UART pointed to in the
		sio structure.
OTHER MODULES:	s_putc
RETURNS:	void
*/
/*
void s_puts(str)
char *str;
{
    for (; *str != NUL; str++)
	s_putc(*str);
}
*/

/*
FUNCTION NAME:	s_puts
LEVEL:		3
LIBRARY:	SIO.LIB
DESCRIPTION:	Transmits a string from the SIO without control through the soctl_
		structure.
OTHER MODULES:	s_putc
RETURNS:	void
*/

s_puts(siop, str)
SIO *siop;
char *str;
{
    while (*str != '\0')
	s_putc(siop, *str++);
}

/*
FUNCTION NAME:	setparity, setdatlen, setstops
LEVEL:		3
LIBRARY:	SIO.LIB
DESCRIPTION:	These functions set their respective data format elements.
OTHER MODULE:	vsetbits
RETURN: 	int: 0 if successful; OR_ARG if argument is out of range;
		BAD_FCN if the function is not supported (SIO pointer is NIL);
		BAD_ARG if the requested setting is not supported by hardware
		(mask initialized to -1).
*/

int setparity(siop, newpar)
SIO *siop;
BYTE newpar;
{
    return (vsetbits(siop, siop->par, newpar));
}

int setdatalen(siop, newdl)
SIO *siop;
BYTE newdl;
{
    return (vsetbits(siop, siop->dl, newdl));
}

int setstops(siop, newsb)
SIO *siop;
BYTE newsb;
{
    return (vsetbits(siop, siop->sb, newsb));
}

/*
FUNCTION NAME:	paritynow, dlnow, stopnow
LEVEL:		3
LIBRARY:	SIO.LIB
DESCRIPTION:	Returns current setting of data format element.
RETURNS:	int: the rank of the setting as defined in SIODEF.H or -1 if
		the function is not supported.
*/

int paritynow(siop)
SIO *siop;
{
    if (siop->par != NIL)
	return (siop->par->now);
    return (-1);
}

int dlnow(siop)
SIO *siop;
{
    if (siop->dl !=NIL)
	return (siop->dl->now);
    return (-1);
}

int stopsnow(siop)
SIO *siop;
{
    if (siop->sb != NIL)
	return (siop->sb->now);
    return (-1);
}

/*
FUNCTION NAME:	setbaud
LEVEL:		3
LIBRARY:	SIO.LIB
DESCRIPTION:	Sets the baud rate to the rank element in the divisor array.
OTHER MODULES:	vsetbr(level-2)
RETURNS:	int: -1 if rank is out of range or not supported by hardware;
		otherwise 0.
COMMENTS:	The rank argument is an index into an array of divisors,
		constants for which reside in SIODEF.H.
*/

int setbaud(siop, rank)
SIO *siop;
unsigned rank;
{
    return (vsetbr(siop, siop->br, rank));
}

/*
FUNCTION NAME:	baudnow
LEVEL:		3
LIBRARY:	SIO.LIB
DESCRIPTION:	Returns rank of current baus rate as define in SIODEF.H.
RETURNS:	int
*/

int baudnow(siop)
SIO *siop;
{
    if (siop->br != NIL)	    /* if supported			*/
	return (siop->br->now);
    return (-1);
}

/*
ARRAY NAME: sbstr, parstr, dlstr, brstr
TYPE:	    Array of char *
LEVEL:	    3
FILE:	    SIO.LIB
DESCRIPTION:ASCII strings for baud rate and data formats.
COMMENTS:   Use the now number of structure to index these arrays.
*/

char *sbstr[]  = {"1", "1 1/2", "2"};
char *parstr[] = {"None", "Odd", "Even", "Mark", "Space"};
char *dlstr[]  = {"5", "6", "7", "8"};
char *brstr[]  = {"50", "75", "110", "134.5", "150", "300", "600", "1200",
		  "1800", "2000", "2400", "3600", "4800", "7200", "9600",
		  "19200", "38400"};

/*
ARRAY NAME: errstr
TYPE:	    Array of char *
LEVEL:	    3
FILE:	    SIO.LIB
DESCRIPTION:Messages for error codes returned by data format and rate functions.
*/

char *errstr[] =
{
    "",
    "Argument out of range",		    /* 1    #define OR_ARG	*/
    "Function not supported",		    /* 2    #define BAD_FCN	*/
    "Argument nor supported",		    /* 3    #define BAD_ARG	*/
    "Illegal device number",		    /* 4    #define BAD_PORT	*/
    "Device not installed",		    /* 5    #define NO_PORT	*/
    "Device not available",		    /* 6    #define NO_SIO	*/
    "Device already open",		    /* 7    #define OPEN_NOW	*/
    "No such device or device not open",    /* 8    #define NO_CLOSE	*/
    "Transfer cancelled from keyboard",     /* 9    #define KEY_CAN	*/
    "File not found",			    /* 10   #define NO_FIL	*/
    ""					    /* NIL array terminator	*/
};

/*
FUNCTION NAME:	setrts, setdtr, setgpo1, setgpo2, setbrk
LEVEL:		3
LIBRARY:	SIO.LIB
DESCRIPTION:	These functions set or reset their respective RS-232 output
		pins.
OTHER MODULES:	vset232
RETURNS:	int: OR_ARG if state argument is other than 0 or 1;
		BAD_FCN if the requested function is not supported in
		hardware; otherwise 0.
*/

int setrts(siop, state) 	/* set RST				*/
SIO *siop;
BOOL state;
{
    return (vset232(siop, siop->rts, state));
}

int setdtr(siop, state) 	/* set DTR				*/
SIO *siop;
BOOL state;
{
    return (vset232(siop, siop->dtr, state));
}

int setgpo1(siop, state)	/* set 1st general-purpose output	*/
SIO *siop;
BOOL state;
{
    return (vset232(siop, siop->gpo1, state));
}

int setgpo2(siop, state)	/* set 2nd general-purpose output	*/
SIO *siop;
BOOL state;
{
    return (vset232(siop, siop->gpo2, state));
}

int setbrk(siop)
SIO *siop;
{
    vset232(siop, siop->brkbit, ON);
    delay(_0_SEC_3);	/* pauses for 300 milliseconds	*/
    //delay(36);		/* 2 seconds for testing	*/
    return (vset232(siop, siop->brkbit,  OFF));
}

/*
FUNCTION NAME:	dtrnow, rtsnow, gpo1now, gpo2now
LEVEL:		3
LIBRARY:	SIO.LIB
DESCRIPTION:	Fetch current states of RS-232 outputs.
OTHER MODULES:	None
RETURNS:	int: BAD_FCN if the output is not supported; otherwise ON/OFF.
*/

int dtrnow(siop)
SIO *siop;
{
    if (siop->dtr == NIL)
	return (BAD_FCN);
    return (siop->dtr->now);
}

int rtsnow(siop)
SIO *siop;
{
    if (siop->rts == NIL)
	return (BAD_FCN);
    return (siop->rts->now);
}

int gpo1now(siop)
SIO *siop;
{
    if (siop->gpo1 == NIL)
	return (BAD_FCN);
    return (siop->gpo1->now);
}

int gpo2now(siop)
SIO *siop;
{
    if (siop->gpo2 == NIL)
	return (BAD_FCN);
    return (siop->gpo2->now);
}

/*
FUNCTION NAME:	iscts, isdsr, isdcd, isri, isgpi1, isgpi2
LEVEL:		3
LIBRARY:	SIO.LIB
DESCRIPTION:	Fetches the state of an RS-232 input.
OTHER MODULES:	vstat232
RETURNS:	int: BAD_FCN if requested function is not supported; otherwise
		TRUE/FALSE.
*/

int iscts(siop)
SIO *siop;		/* pointer to the serial device structure	*/
{
    return (vstat232(siop, siop->cts));
}

int isdsr(siop)
SIO *siop;
{
    return (vstat232(siop, siop->dsr));
}

int isdcd(siop)
SIO *siop;
{
    return (vstat232(siop, siop->dcd));
}

int isri(siop)
SIO *siop;
{
    return (vstat232(siop, siop->ri));
}

int isgpi1(siop)
SIO *siop;
{
    return (vstat232(siop, siop->gpi1));
}

int isgpi2(siop)
SIO *siop;
{
    return (vstat232(siop, siop->gpi2));
}

/*
FUNCTION NAME:	xmiton, rcvon, txhson, rxhson, setclk
LEVEL:		3
LIBRARY:	SIO.LIB
DESCTIPTION:	Enable transmitter and receiver; enable reansmitter handshake
		on CTS and receiver handshake on DCD; set clock factor.
OTHER MODULES:	vsetbits
RETURNS:	int: OR_ARG if argument is out of renge; BAD_FCN if the
		function is not supported(SIO pointer is NIL);
		BAD_ARG if the requested setting is not supported by hardware;
		0 if successful.
*/

int xmiton(siop, state) 	/* transmitter on/off			*/
SIO *siop;
int state;
{
    return(vsetbits(siop, siop->txenable, state));
}

int rcvon(siop, state)		/* receiver on/off			*/
SIO *siop;
int state;
{
    return(vsetbits(siop, siop->rxenable, state));
}

int txhson(siop, state) 	/* transmitter handshake on/off 	*/
SIO *siop;
int state;
{
    return(vsetbits(siop, siop->ctshs, state));
}

int rxhson(siop, state) 	/* receiver handshake on/off		*/
SIO *siop;
int state;
{
    return(vsetbits(siop, siop->dcdhs, state));
}

int setclk(siop, clkfactor)	/* set clock factor			*/
SIO *siop;
int clkfactor;
{
    return(vsetbits(siop, siop->clkf, clkfactor));
}

/*
FUNCTION NAME:	s_open
LEVEL:		3
LIBRARY:	SIO.LIB
DESCRIPTION:	"Opens" an SIO.
OTHER MODULES:	setdtr, setrts, setgpo1, setgpo2, setparity, setstops,
		setdatalen, setbaud, xmiton, rcvon, rxhson, txhson, s_config.
RETURNS:	SIO *; NIL if no stroage is available, if the portnum
		argument is out of renge, or if the portnum is no defined.
		Error code for the failure is placed in the global varable
		s_errno.
COMMENTS:	The SIO (serial port) described is initialized first by
		calling s_config to perform any UART-specific tasks, then
		by calling setbaud, setparity, ect. A pointer to this SIO
		is returned.
*/
int s_errno;			/* GLOBAL VARIABLE			*/

SIO *s_open(portnum)
unsigned portnum;		/* must be unsigned			*/
{
    extern struct _siostat  _siolist[];
    extern int	_maxsios;		/* global -- number of ports	*/
    SIO *siop;				/* @[SIO]@ pointer to return	*/
    if (portnum >= _maxsios)		/* is device number legal?	*/
	{
	s_errno = BAD_PORT;
	return (NIL);
	}
/*
    if (isport(portnum) == NUL) 	Is the hardware installed?
	{				This function has desabled,
	s_errno = NO_PORT;		because this is a special
	return (NIL);			multi serial-ports library.
	}				Function isport containd in IBMPC.C.
*/
    if (_siolist[portnum].typep == NIL) /* does an SIO exit for it?	*/
	{
	s_errno = NO_PORT;
	return (NIL);
	}
    if (_siolist[portnum].openflag != CLOSED)	/* is it already open?	*/
	{
	s_errno = OPEN_NOW;
	return (NIL);
	}
    siop = _siolist[portnum].typep;	/* SIO to intermediate pointer	*/
    s_config(siop);			/* setup SIO's uart             */
    setclk(siop, siop->clkf->start);	/* set clock factor		*/
    setparity(siop, siop->par->start);	/* set data format...		*/
    setstops(siop, siop->sb->start);
    setdatalen(siop, siop->dl->start);
    setbaud(siop, siop->br->start);
    setdtr(siop, siop->dtr->start);	/* set RS-232 and GP outputs	*/
    setrts(siop, siop->rts->start);
    setgpo1(siop, siop->gpo1->start);
    setgpo2(siop, siop->gpo2->start);
    rxhson(siop, siop->dcdhs->start);	/* set receiver handshaking	*/
    txhson(siop, siop->ctshs->start);	/* set xmitter handshaking	*/
    xmiton(siop, siop->txenable->start); /* enable transmitter		*/
    rcvon(siop, siop->rxenable->start); /* enable receiver		*/
    _siolist[portnum].openflag = OPEN;	/* mark SIO open in list	*/
    siop->devnum = portnum;		/* put device number in SIO	*/

    return (siop);
}

/*
FUNCTION NAME:	s_close
LEVEL:		3
LIBRARY:	SIO.LIB
DESCRIPTION:	Restores the SIO pointed to by siop to its original
		condition.
OTHER MODULES:	s_restore
RETURNS:	int: NO_CLOSE if the SIO pointer is NIL or nor on the list
		of s_open'ed ports.
COMMENTS:	If the close fails, no attempt is made to restore the SIO.
*/

int s_close(siop)
SIO *siop;
{
    extern struct _siostat  _siolist[];     /* list of available @[SIO]@s*/
    extern int _maxsios;    /* global -- maximum number of serial ports */
    int i;
    if (siop == NIL)
	return (NO_CLOSE);
    for (i = 0; i < _maxsios; i++)
	if ((siop == _siolist[i].typep) && (_siolist[i].openflag == OPEN))
	    {
	    _siolist[i].openflag = CLOSED;
	    break;
	    }
    if (i == _maxsios)		    /* no match in list 		*/
	return (NO_CLOSE);

    s_restore(siop);		    /* restore interface		*/
    siop->devnum = VIRGIN;	    /* install dummy device number	*/
    return (0);
}

/*
FUNCTION NAME:	s_keybort
LEVEL:		3
LIBRARY:	SIO.LIB
DESCRIPTION:	Checks the keyboard to see if the designated abort key has
		been typed.
OTHER MODULES:	inkey
RETURNS:	BOOL: TRUE if the abort key has been pressed; otherwise FALSE.
COMMENTS:	The abort key is defined by the kbortc member of the soctl_
		structure. This function can be replaced with the following
		macro:
		#define s_keybort(p)	inkey()==(p)->s_octl->kbortc
*/

BOOL s_keybort(siop)
SIO *siop;
{
    return (inkey() == siop->s_octl->kbortc);
}

/*
FUNCTION NAME:	s_fputs
LEVEL:		3
LIBRARY:	SIO.LIB
DESCRIPTION:	Outputs a string to the SIO under control of  the soctl_
		structure.
OTHER MODULES:	s_fputc
RETURNS:	void
COMMENTS:	The local function s_fflush clears s_fputc's single-character
		buffer.
*/

static void s_fflush(siop)
SIO *siop;
{
    if (siop->s_octl->flushflg && !siop->s_octl->rawflg)
	s_fputc(siop, siop->s_octl->lastc);
}

void s_fputs(siop, str)
SIO *siop;
char *str;
{
    while (*str != '\0')
	s_fputc(siop, *str++);
    s_fflush(siop);
    if (siop->s_octl->consflg == TRUE)
	putchar('\n');
}

/*
FUNCTION NAME:	s_printf
LEVEL:		3
LIBRARY:	SIO.LIB
DESCRIPTION:	"Printf-style" output through the SIO.
OTHER MODULES:	s_fputc and your compiler's version of format.
RETURNS:	int: the number of characters output.
COMMENTS:	Depending upon the internal construction of fotmat, the return
		value may be affected by the return value of s_fputc.
*/

//static SIO *dummyp;	/* local for passing pointer to format function */
//static void dummy(c)
//int c;
//{
//    s_fputc(dummyp, c);
//}
//
//int s_printf(siop, fmtstr, argp)
//SIO *siop;
//char *fmtstr;
//char argp;	/* must be char     */
//{
//    dummyp = siop;	/* pass the SIO pointer in a dummy local	*/
//    format(dummy, fmtstr, &argp);
//    s_fflush(siop);
//}

/*
FUNCTION NAME:	s_ocntrl
LEVEL:		3
LIBRARYL	SIO.LIB
DESCRIPTION:	Sets or fetches the value of the flag, delay, and mode
		members of an sictl_ input control structure.
OTHER MODULES:	None
RETURNS:	int.  SET: 0 if successful, otherwise -1; GET: -1 if
		unsuccessful.
COMMENTS:	Constants are in SIOCTL.H. The range of the 'code' arguments
		is not checked.
*/

int s_ocntrl(siop, function, member, newval)
SIO	    *siop;
short	    function;	    /* SET or GET				*/
short	    member;	    /* code for member to address		*/
unsigned    newval;	    /* new value for member (SET only)		*/
{
    int retval = 0;
    if (function == SET)
	switch (member)
	    {
	    case RAWFLAG:	siop->s_octl->rawflg	= newval; break;
	    case EOLMODE:	siop->s_octl->eolmode	= newval; break;
	    case CONSFLAG:	siop->s_octl->consflg	= newval; break;
	    case WECHOFLAG:	siop->s_octl->wechoflg	= newval; break;
	    case WECHO_DLY:	siop->s_octl->wecho_dly = newval; break;
	    case CASEMODE:	siop->s_octl->casemode	= newval; break;
	    case XOFFC: 	siop->s_octl->xoffc	= newval; break;
	    case XONC:		siop->s_octl->xonc	= newval; break;
	    case XXMODE:	siop->s_octl->xxmode	= newval; break;
	    case ASCIIFLAG:	siop->s_octl->asciiflg	= newval; break;
	    case CNTRLFLAG:	siop->s_octl->cntrlflg	= newval; break;
	    case FILTERFLAG:	siop->s_octl->filterflg = newval; break;
	    case IB_DLYFLAG:	siop->s_octl->ib_dlyflg = newval; break;
	    case IB_DLY:	siop->s_octl->ib_dly	= newval; break;
	    case IL_DLYFLAG:	siop->s_octl->il_dlyflg = newval; break;
	    case IL_DLY:	siop->s_octl->il_dly	= newval; break;
	    case IL_DLYC:	siop->s_octl->il_dlyc	= newval; break;
	    case KBORTC:	siop->s_octl->kbortc	= newval; break;
	    default:
		retval = -1;
		break;
	    }
    else if (function == GET)
	switch (member)
	    {
	    case RAWFLAG:	retval = siop->s_octl->rawflg;	    break;
	    case EOLMODE:	retval = siop->s_octl->eolmode;     break;
	    case CONSFLAG:	retval = siop->s_octl->consflg;     break;
	    case WECHOFLAG:	retval = siop->s_octl->wechoflg;    break;
	    case WECHO_DLY:	retval = siop->s_octl->wecho_dly;   break;
	    case CASEMODE:	retval = siop->s_octl->casemode;    break;
	    case XOFFC: 	retval = siop->s_octl->xoffc;	    break;
	    case XONC:		retval = siop->s_octl->xonc;	    break;
	    case XXMODE:	retval = siop->s_octl->xxmode;	    break;
	    case ASCIIFLAG:	retval = siop->s_octl->asciiflg;    break;
	    case CNTRLFLAG:	retval = siop->s_octl->cntrlflg;    break;
	    case FILTERFLAG:	retval = siop->s_octl->filterflg;   break;
	    case IB_DLYFLAG:	retval = siop->s_octl->ib_dlyflg;   break;
	    case IB_DLY:	retval = siop->s_octl->ib_dly;	    break;
	    case IL_DLYFLAG:	retval = siop->s_octl->il_dlyflg;   break;
	    case IL_DLY:	retval = siop->s_octl->il_dly;	    break;
	    case IL_DLYC:	retval = siop->s_octl->il_dlyc;     break;
	    case KBORTC:	retval = siop->s_octl->kbortc;	    break;
	    default:
		retval = -1;
		break;
	    }
    return (retval);
}

/*
FUNCTION NAME:	s_opush, s_opop
LEVEL:		3
LIBRARY:	SIO.LIB
DESCRIPTION:	Save and restore s_octl structures on local stack.
OTHER MODULES:	None
RETURNS:	int: 0 if stack operation is successful; otherwise 1.
COMMENTS:	Both functions check for blance of the local stack, so take
		care that every "push" is blanced with a corresponding "pop".
*/

#define OSTKSIZ     10
static struct soctl_ *ostack[OSTKSIZ] = {NIL};	/* init lst element to NIL*/
static struct soctl_ **ostackp = ostack;	/* pointer to top of stack*/

int s_opush(siop)
SIO *siop;
{
    char *malloc();
    if (ostackp >= &ostack[OSTKSIZ])	/* check for end of array	*/
	{
	puts("\007Ostack overflow.");
	return (1);
	}
    *ostackp = (struct soctl_ *)malloc(sizeof(struct soctl_));	/* get mem*/
    if (*ostackp == NIL)	/* out of memory			*/
	return (1);
    **ostackp = *siop->s_octl;	/* make copy in allocated memory	*/
    ++ostackp;			/* point to next slot			*/
    return (0);
}

int s_opop(siop)
SIO *siop;
{
    if (ostackp <= ostack)	/* don't let stack become unblanced     */
	{
	puts("\007Istack underflow.");
	return (1);
	}
    --ostackp;			/* point at first active element	*/
    *siop->s_octl = **ostackp;	/* restore old structure from stack	*/
    free(*ostackp);		/* release allocated memory		*/
    return (0);
}

/*
FUNCTION NAME:	s_icntrl
LEVEL:		3
LIBRARY:	SIO.LIB
DESCRIPTION:	Sets or fetches the value of the flag, delay, and mode
		members of an sictl_ input control structure.
OTHER MODULES:	None
RETURNS:	int: 0 if successful; otherwise, -1.
COMMENTS:	Constants are in SIOCTL.H. The range of the 'code' agruments
		is not checked.
*/

int s_icntrl(siop, function, member, newval)
SIO *siop;
short function; 		/* SET or GET (see SIOCTL.H)		*/
short member;			/* code for member to address		*/
unsigned newval;		/* new value for member (SET only)	*/
{
    int retval = 0;
    if (function == SET)
	switch (member)
	    {
	    case T_OUT: 	siop->s_ictl->t_out	= newval;   break;
	    case T_OUTFLAG:	siop->s_ictl->t_outflg	= newval;   break;
	    case RAWFLAG:	siop->s_ictl->rawflg	= newval;   break;
	    case EKOBAKFLAG:	siop->s_ictl->ekobakflg = newval;   break;
	    case ASCIIFLAG:	siop->s_ictl->asciiflg	= newval;   break;
	    case EOLMODE:	siop->s_ictl->eolmode	= newval;   break;
	    case CONSFLAG:	siop->s_ictl->consflg	= newval;   break;
	    case CASEMODE:	siop->s_ictl->casemode	= newval;   break;
	    case XONC:		siop->s_ictl->xonc	= newval;   break;
	    case XOFFC: 	siop->s_ictl->xoffc	= newval;   break;
	    case CNTRLFLAG:	siop->s_ictl->cntrlflg	= newval;   break;
	    case FILTERFLAG:	siop->s_ictl->filterflg = newval;   break;
	    default:
		retval = -1;
	    }
    else if (function == GET)
	switch (member)
	    {
	    case T_OUT: 	retval = siop->s_ictl->t_out;	    break;
	    case T_OUTFLAG:	retval = siop->s_ictl->t_outflg;    break;
	    case RAWFLAG:	retval = siop->s_ictl->rawflg;	    break;
	    case EKOBAKFLAG:	retval = siop->s_ictl->ekobakflg;   break;
	    case ASCIIFLAG:	retval = siop->s_ictl->asciiflg;    break;
	    case EOLMODE:	retval = siop->s_ictl->eolmode;     break;
	    case CONSFLAG:	retval = siop->s_ictl->consflg;     break;
	    case CASEMODE:	retval = siop->s_ictl->casemode;    break;
	    case XONC:		retval = siop->s_ictl->xonc;	    break;
	    case XOFFC: 	retval = siop->s_ictl->xoffc;	    break;
	    case CNTRLFLAG:	retval = siop->s_ictl->cntrlflg;    break;
	    case FILTERFLAG:	retval = siop->s_ictl->filterflg;   break;
	    default:
		retval = -1;
	    }
    return (retval);
}

/*
FUNCTION NAME:	s_ipush, s_ipop
LEVEL:		3
LIBRARY:	SIO.LIB
DESCRIPTION:	Save and restore s_ictl structures on the local stack.
OTHER MODULES:	None
RETURNS:	int: 0 if stack operation is successful; otherwise 1.
COMMENTS:	Both functions check for balance of the local stack, so
		take care that erery "push" is balance with a corresponding
		"pop".
*/

#define ISTKSIZ 10
static struct sictl_ *istack[ISTKSIZ] = {NIL};	/* init lst element to NIL*/
static struct sictl_ **istackp = istack;	/* pointer to top of stack*/

int s_ipush(siop)
SIO *siop;
{
    char *malloc();
    if (istackp >= &istack[ISTKSIZ])
	{
	puts("\007Istack overflow.");
	return (1);
	}
    *istackp = (struct sictl_ *)malloc(sizeof(struct sictl_));
    if (*istackp == NIL)
	return (1);
    **istackp = *siop->s_ictl;
    ++istackp;
    return (0);
}

int s_ipop(siop)
SIO *siop;
{
    if (istackp <= istack)
	{
	puts("\007Istack underflow.");
	return (1);
	}
    --istackp;
    *siop->s_ictl = **istackp;
    free(*istackp);
    return (0);
}

/*
FUNCTION NAME:	s_fgets
LEVEL		3
LIBRARY:	SIO.LIB
DESCRIPTION:	Inputs a NUL-terminated string of numc bytes into buffer.
OTHER MODULES:	s_fgets, s_iflush.
RETURNS:	int: number of bytes in buffer (0 if timeout occurs).
COMMENTS:	Buffer must be numc+1 bytes long to accommodate the NUL
		terminator. Timeout mode is forced for the duration of the
		function.
*/

int s_fgets(siop, buff, numc)
SIO *siop;
BYTE *buff;
int numc;
{
    int c, count;
    s_ipush(siop);			    /* save input structure	*/
    s_icntrl(siop, SET, T_OUTFLAG, ON);     /* turn on input timeout	*/
    if (s_icntrl(siop, GET, T_OUT) == 0)    /* save timeout interval	*/
	s_icntrl(siop, SET, T_OUT, 1);	    /* make sure there is some delay*/
    count = 0;
    while (count < numc)
	{
	if ((c = s_fgetc(siop)) == TIMEOUT)
	    break;
	if (c == IGNORE)		/* exclude this character	*/
	    continue;
	*buff++ = c;			/* install byte in buffer	*/
	++count;
	}
    if ((c = s_iflush(siop)) != -1)	/* byte pending in s_fgetc?	*/
	{
	*buff++ = c;
	++count;
	}
    *buff = NUL;			/* terminate buffer with NUL	*/
    s_ipop(siop);			/* save input structure 	*/
    return (count);
}

/*
FUNCTION NAME:	s_fgetln
LEVEL:		3
LIBRARY:	SIO.LIB
DESCRIPTION:	Inputs a NUL-terminated string of bytes into buff. Input is
		terminated when the specified delimiter is encountered.
OTHER MODULES:	s_fgetc, s_iflush.
RETURNS:	int: number of bytes in buffer (0 if timeout occurs).
COMMENTS:	Buffer must be numc+1 bytes long to accommodate the NUL
		terminator. Timout mode is forced for the duration of the
		function.
*/

int s_fgetln(siop, buff, numc, eolc)
SIO *siop;
BYTE *buff;
int numc;
BYTE eolc;
{
    int c, count;
    s_ipush(siop);			/* save input structure 	*/
    s_icntrl(siop, SET, T_OUTFLAG, ON);	/* trun on input timeout	*/
    if (s_icntrl(siop, GET, T_OUT) == 0) /* save timeout interval	*/
	s_icntrl(siop, SET, T_OUT, 1);	/* make sure there is some delay*/
    count = 0;
    while (count < numc)
	{
	if ((c = s_fgetc(siop)) == TIMEOUT)
	    break;
	if (c == IGNORE)		/* exclude this character	*/
	    continue;
	if (c == eolc)			/* don't put eol byte in buffer */
	    break;
	*buff++ = c;			/* install byte in buffer	*/
	++count;
	}
    if ((c = s_iflush(siop)) != -1)	/* byte pending in s_fgetc?	*/
	{
	*buff++ = c;
	++count;
	}
    *buff = NUL;
    s_ipop(siop);			/* restore input structure	*/
    return (count);
}

/*
FUNCTION NAME:	s_clrsio
LEVEL:		3
LIBRARY:	SIO.LIB
DESCRIPTION:	Clears numbytes byter from UART's receive buffers.
OTHER MODULES:	s_waitch
RETURNS:	TIMEOUT(-1) if no bytes received, or 0 after numbytes bytes
		are received.
COMMENTS:	To avoid inaccuracies inherent in the timing functions, a
		minimum delay of _0_SEC_1 must be used.
*/

int s_clrsio(siop, numbytes)
SIO *siop;
unsigned numbytes;
{
    for (; numbytes > 0; --numbytes)
	if (s_waitch(siop, _0_SEC_1) == TIMEOUT)
	    return (TIMEOUT);
    return (0);
}
