
/****************************************************************************

    Level 1 library for UART 8250.

    This library contains following functions,

	s_rcv	    Fetch a byte from the UART's receiver.
	s_rcvstat   Return the status of the UART's receiver.
	s_xmit	    Dispatch a byte from the UART's transmitter.
	s_xmitstat  Return the status of the UART's transmitter.
	_vsetbits   This is the primitive to write the format in the virtual
		    register to the corresponding UART register.
	_vsetbr     Primitive for setting the baud rate.
	s_config    Copies the contents of the 8250's write registers into
		    the SIO's virtual array.
	s_restore   Restores the SIO to its original state.
	_vset232    Writes the virtual register containing the selected
		    RS-232 output pin to the corresponding UART register.
	_vstat232   Reads the UART register containing the status bit for
		    the appropriate RS-232 input. A pointer to this function
		    resides in each vin232_ structure.

****************************************************************************/

#include    "stdio.h"
#include    "siodef.h"
#include    "u8250.h"

/*
FUNCTION NAME:	s_rcv, s_rcvstat, s_xmit, s_xmitstat
LEVEL:		1
LIBRARY:	U8250.C and Z80SIO.C
DESCRIPTION:	UART I/O primitives written to access their data though a
		pointer to an SIO(data structure).
OTHER MOUDLS:	None
COMMENTS:	Caution! calling these functions without the pointer argument
		produces disastrous results.
*/

BYTE s_rcv(siop)
register SIO *siop;
{
    return ((*siop->readbyte)(siop->uart_base + siop->data_off));
}

BYTE s_rcvstat(siop)
register SIO *siop;
{
    return ((*siop->readbyte)(siop->uart_base + siop->status_off) & siop->rcvmask);
}

void s_xmit(siop, c)
register SIO *siop;
BYTE c;
{
    (*siop->writebyte)(siop->uart_base + siop->data_off, c);
}

BYTE s_xmitstat(siop)
register SIO *siop;
{
    return ((*siop->readbyte)(siop->uart_base + siop->status_off) & siop->xmitmask);
}

/*
FUNCTION NAME:	_vsetbits
LEVEL:		1
LIBRARY:	U8250.LIB
DESCRIPTION:	This is the primitive to write the format in the virtual to
		the corresponding UART register.
OTHER MODULES:	None
RETURNS:	void
COMMENTS:	This function is called only by vsetbits.
*/

void _vsetbits(siop, brp, binval)
register SIO *siop;
struct vregbits_ *brp;
BYTE binval;
{
    (*siop->writebyte)(siop->uart_base + brp->offset, binval);
}

/*
FUNCTION NAME:	_vsetbr
LEVEL:		1
LIBRARY:	U8250.LIB
DESCRIPTION:	Primitive foe setting the baud rate.
OTHER MODULES:	None
RETURNS:	viod
COMMENTS:	The baud rate is changed by writing lobaud and hibaud to
		the 8250's LSB and MSB Baud registers. These registers can
		be addressed only if bit 7 in the Data Format register is
		TURE. After writting to these two registers, the orignal
		contents of the Data Format
		register are restored.
*/

void _vsetbr(siop, brp, lobaud, hibaud)
register SIO *siop;
struct vbaud_ *brp;
BYTE lobaud;
BYTE hibaud;
{
    unsigned tmpfmt;
    tmpfmt = (*siop->readbyte)(siop->uart_base + siop->par->offset) | 0x80;
    (*siop->writebyte)(siop->uart_base + siop->par->offset, tmpfmt);
    (*siop->writebyte)(siop->uart_base + brp->offsetlo, lobaud);
    (*siop->writebyte)(siop->uart_base + brp->offsethi, hibaud);
    (*siop->writebyte)(siop->uart_base + siop->par->offset, tmpfmt & 0x7f);
}

/*
FUNCTION NAME:	s_config
LEVEL:		1
LIBRARY:	U8250.LIB
DESCRIPTION:	Copies the contents of the 8250's write registers into the
		SIO's virtual array.
OTHER MODULES:	None
RETURNS:	void
*/

void s_config(siop)	    /* copy physical UART regs into virtual regs    */
register SIO *siop;
{
    BYTE tmp;
    int i;
    tmp = (*siop->readbyte)(siop->uart_base + siop->par->offset) | 0x80;
    (*siop->writebyte)(siop->uart_base + siop->par->offset, tmp);
    siop->v_regp[VBAUDLO][USR] = (*siop->readbyte)(siop->uart_base + siop->br->offsetlo);
    siop->v_regp[VBAUDHI][USR] = (*siop->readbyte)(siop->uart_base + siop->br->offsethi);
    (*siop->writebyte)(siop->uart_base + siop->par->offset, tmp & 0x7f);
    siop->v_regp[VINTR][USR] = (*siop->readbyte)(siop->uart_base + 1);
    siop->v_regp[VFMT][USR] = (*siop->readbyte)(siop->uart_base + siop->par->offset);
    siop->v_regp[V232OUT][USR] = (*siop->readbyte)(siop->uart_base + 4);
    for (i=0; i<5; i++)     /* copy USR column into VIR column	    */
	siop->v_regp[i][VIR] = siop->v_regp[i][USR];
}

/*
FUNCTION NAME:	s_restore
LEVEL:		1
LIBRARY:	U8250.LIB
DESCRIPTION:	Restores the SIO to its original state.
OTHER MODULES:	None
RETURNS:	void
*/

void s_restore(siop)
register SIO *siop;
{
    _vsetbr(siop, siop->br, siop->v_regp[VBAUDLO][USR], siop->v_regp[VBAUDHI][USR]);
    _vsetbits(siop, siop->par, siop->v_regp[VFMT][USR]);
    (*siop->writebyte)(siop->uart_base + 1, siop->v_regp[VINTR][USR]);
    (*siop->writebyte)(siop->uart_base + 4, siop->v_regp[V232OUT][USR]);
}

/*
FUNCTION NAME:	_vset232
LEVEL:		1
LIBRARY:	U8250.LIB
DESCRIPTION:	Writes the virtual register containing the selected RS-232
		output pin to the corresponding UART register.
RETURNS:	void
COMMENTS:	This function is called only by vset232.
*/

void _vset232(siop, p232, binval)
register SIO *siop;
struct vout232_ *p232;
BYTE binval;
{
    (*siop->writebyte)(siop->uart_base + p232->offset, binval);
}

/*
FUNCTION NAME:	_vstat232
LEVEL:		1
LIBRARY:	U8250.LIB
DESCRIPTION:	Reads the UART register containing the status bit for the
		appropriate RS-232 input. A pointer to this function resides
		in each vin232_ structure.
RETURNS:	BYTE
*/

BYTE _vstat232(siop, p232)
register SIO *siop;
struct vin232_ *p232;
{
    return ((*siop->readbyte)(siop->uart_base + p232->offset));
}
