
/***************************************************************************

    Library for XMODEM file transfers

    This library contains follow functions and arrays:

	x_msg		Array of pointers to message strings used in XMODEM
			file transfer modules.

	x_except	Handles exceptions(e.g., errors) that occur during
			XMODEM transfers.

	makepacket	Makes sndpackets in buffer. Check value character is
			not installed.

	ckvinstall	Installs check value(s) in the array of packet
			structures.

	x_snd		Transmits a file using XMODEM file-transfer protocol.

	x_sndcksum	Calculates and returns an arithmetic checksum on the
			buffer indicated.

	x_rcvcksum	Verifies 8-bits arithmetic checksum on indicated
			buffer.

	polydiv 	Simulates longhand polynomial division.

	x_sndcrc	Calculates CRC for indicated XMODEM data block.

	x_rcvcrc	Calculates CRC for an rcvpacket.

***************************************************************************/

#include    "ascii.h"
#include    <stdio.h>
#include    "siodef.h"
#include    "sioctl.h"
#include    "xmod.h"
#include    "crc.h"

/* GLOBAL variables for XMODEM.LIB  */
int paksize;			/* number of bytes to xmit or rcv	*/
unsigned short (*r_errckp)();	/* pointer to the errir check function	*/
char *ckvname;			/* string indicating CRC or cksum	*/
BYTE csync;			/* NAK for checksum, 'C' for crc	*/
unsigned short *crctblp;	/* pointer to the CRC loopup table	*/
struct rcvpacket *rcvbuff;	/* pointer to receive packet buffer	*/

/*
ARRAY NAME:	x_msg
TYPE:		char *
LIBRARY:	XMODEM.LIB
DESCRIPTION:	Array of pointers to message strings used in XMODEM file
		transfer modules.
COMMENTS:	These rank-equivalent constants are shown in comments.
*/

char *x_msg[] =
{
    "\nTransfer cancelled from keyboard.\007",		/* E_USRCAN	0   */
    "Input file is empty\007",				/* E_FILEMTY	1   */
    "Timeout waiting for ACK",				/* E_NOACK	2   */
    "Receiver cancelled transfer.\007", 		/* E_RCVCAN	3   */
    "Resending last packet",				/* E_BADPAK	4   */
    "End of file read", 				/* E_EOF	5   */
    "\nWarning: receiver did not acknowledge EOT.\007", /* E_LASTACK	6   */
    "Successful transmission",				/* E_SNDOK	7   */
    "Fatal error reading disk.",			/* E_DSKREAD	8   */
    "Timeout waiting for SOH",				/* E_NOSOH	9   */
    "\nData error detected. Requesting retransmission.",/* E_BADCKV	10  */
    "Transfer cancelled by sender.\007",		/* E_SNDCAN	11  */
    "\nByte received, but not SOH",			/* E_BADSOH	12  */
    "\nTimeout waiting for packet data",		/* E_NODATA	13  */
    "\nBad packet number",				/* E_PAKNUM	14  */
    "\nSender missed ACK",				/* E_SNDACK	15  */
    "Fatal packet sequence error.\007", 		/* E_PAKSEQ	16  */
    "Fatal error writing disk.",			/* E_DSKWRITE	17  */
    "Successful reception"				/* E_RCVOK	18  */
};

/*
FUNCTION NAME:	x_except
LEVEL:		3
LIBRARY:	XMODEM.LIB
DESCRIPTION:	Handles exceptions(e.g.,errors) that occur during XMODEM
		transfers.
RETURNS:	int: one of the error codes given in XMOD.H.
COMMENTS:	A pointer ro an exception counter variable and its maximum
		permitted value are passed as arguments.
*/

int x_except(siop, exceptnum, ex_cntp, maxexcept)
SIO *siop;
int exceptnum;		    /* error code in XMOD.H			*/
int *ex_cntp;		    /* pointer to an error counter		*/
int *maxexcept; 	    /* max number of errors permitted		*/
{
    extern unsigned short x_rcvcrc(), x_rcvcksum();
    int errval = CONTINUE;
    switch (exceptnum)
	{
	case E_USRCAN:	    /* user aborts from keyboard		*/
	    puts(x_msg[E_USRCAN]);
	errval = BREAK;
	delay(_0_SEC_5);
	s_putc(siop, CAN);
	break;

	/* errors in transmission	*/
	case E_FILEMTY:
	    puts(x_msg[E_FILEMTY]);
	    break;
	case E_RCVCAN:		    /* transmitter aborted		*/
	    puts(x_msg[E_RCVCAN]);
	    errval = BREAK;	    /* FATAL				*/
	    break;
	case E_NOACK:		    /* timeout waiting for ACK		*/
	    puts(x_msg[E_NOACK]);
	    ++*ex_cntp;
	    break;
	case E_BADPAK:		    /* resending previous packet	*/
	    puts(x_msg[E_BADPAK]);
	    ++*ex_cntp;
	    s_clrsio(siop, CRCPAKSIZ);	/* clear buffer 		*/
	    break;
	case E_EOF:		    /* end of disk file read		*/
	    puts(x_msg[E_EOF]);
	    errval = E_EOF;
	    *ex_cntp = X_ESNDMAX;
	    break;
	case E_DSKREAD: 	    /* disk read error			*/
	    puts(x_msg[E_DSKREAD]);
	    errval = BREAK;	    /* FATAL				*/
	    break;
	case E_LASTACK: 	    /* no ACK response to EOT		*/
	    puts(x_msg[E_LASTACK]);
	    s_putc(siop, CAN);
	    break;
	case E_SNDOK:		    /* good transfer			*/
	    errval = E_EOF;
	    printf("%s", x_msg[E_SNDOK]);   /* use printf -- no new line*/
	    break;

	/* errors in reception	*/
	case E_NOSOH:		    /* timeout waiting for SOH		*/
	    ++*ex_cntp; 	    /* bump caller's error count        */
	    puts(x_msg[E_NOSOH]);
	    if (rcvbuff->pnum1 == VIRGIN)   /* negotiate startup mode	*/
		{
		r_errckp = (r_errckp == x_rcvcrc) ? x_rcvcksum	: x_rcvcrc;
		paksize  = (r_errckp == x_rcvcrc) ? CRCPAKSIZ-1 : CKSPAKSIZ-1;
		csync	 = (r_errckp == x_rcvcrc) ? 'C' 	: NAK;
		ckvname  = (r_errckp == x_rcvcrc) ? "CRC"	: "CHECKSUM";
		s_putc(siop, csync);
		printf("\nAttempting to synchronize in %s mode.\n", ckvname);
		}
	    else
		s_putc(siop, NAK);
	    break;
	case E_BADCKV:		    /* bad checksum or CRC		*/
	    puts(x_msg[E_BADCKV]);
	    ++*ex_cntp;
	    s_putc(siop, NAK);
	    break;
	case E_SNDCAN:		    /* transmitter aborted		*/
	    puts(x_msg[E_SNDCAN]);
	    errval = BREAK;	    /* FATAL				*/
	    break;
	case E_BADSOH:		    /* invalid SOH received		*/
	    puts(x_msg[E_BADSOH]);
	    s_clrsio(siop, CRCPAKSIZ);	/* ignore reset of packet	*/
	    s_putc(siop, NAK);	    /* tell sender to retry		*/
	    ++*ex_cntp;
	    break;
	case E_NODATA:		    /* timeout in packet		*/
	    puts(x_msg[E_NODATA]);
	    ++*ex_cntp;
	    s_putc(siop, NAK);
	    break;
	case E_PAKNUM:		    /* packet numbers don't agree       */
	    puts(x_msg[E_PAKNUM]);
	    s_clrsio(siop, CRCPAKSIZ);
	    ++*ex_cntp;
	    s_putc(siop, NAK);	    /* ask for retransmission		*/
	    break;
	case E_SNDACK:		    /* duplicate of previous packet	*/
	    puts(x_msg[E_SNDACK]);
	    ++*ex_cntp;
	    s_putc(siop, ACK);
	    break;
	case E_PAKSEQ:		    /* packets out of order -- FATAL	*/
	    puts(x_msg[E_PAKSEQ]);
	    errval = BREAK;	    /* FATAL				*/
	    break;
	case E_DSKWRITE:	    /* disk write error 		*/
	    puts(x_msg[E_DSKWRITE]);
	    errval = BREAK;	    /* FATAL				*/
	    s_putc(siop, CAN);
	    break;
	case E_RCVOK:		    /* good transfer			*/
	    errval = BREAK;
	    printf("%s", x_msg[E_RCVOK]);   /* use printf -- no newline */
	    delay(_0_SEC_5);	    /* let receiver get ready		*/
	    s_putc(siop, ACK);	    /* send final ACK			*/
	    break;
	default:
	    break;
	}
    if (errval == CONTINUE && *ex_cntp > *maxexcept)
	{
	errval = BREAK;
	printf("%d errors or waits have occured: Keep trying?  (y/n):\007",
		maxexcept);
	if (toupper(getch()) == 'Y')
	    {
	    errval = CONTINUE;
	    *ex_cntp = 0;	    /* reset caller's error counter     */
	    }
	putchar('\n');
	}
    return (errval);
}

/*
FUNCTION NAME:	makepacket
LEVEL:		3
LIBRARY:	XMODEM.LIB
DESCRIPTION:	Makes sndpackets in buffer. Checkvalue character is not
		installed.
OTHER MODULES:	None
RETURNS:	unsigned int: the number of packets made.
COMMENTS:	The checkvalues cannot be installed until after start-up.
*/

unsigned int makepacket(pbp, pakcnt, paknum, fp)
struct	    sndpacket *pbp;	/* "packet buffer pointer"		*/
unsigned    int pakcnt; 	/* number of packets in buffer		*/
unsigned    long paknum;	/* current packet count 		*/
FILE	    *fp;
{
    unsigned paksread = 0;	/* first read in data from disk 	*/
    unsigned byteread;
    unsigned buffsize;
    BYTE endfile = 0X1A;
    buffsize = sizeof(pbp->data);
    while ((byteread = fread(pbp->data, sizeof(BYTE), buffsize, fp)) != 0)
	{
	pbp->soh    = SOH;	/* install SOH				*/
	pbp->pnum1  = paknum++ % 256;	/* install packet number and... */
	pbp->pnum2  = ~(pbp->pnum1);	/* its one's complement         */
	while (byteread < buffsize)	/* fill (0x01A) to remainder	*/
	    {
	    pbp->data[byteread] = endfile;
	    byteread++;
	    }

	if (++paksread == pakcnt)
	    break;
	++pbp;			/* next packet, please			*/
	}
    return (paksread);
}

/*
FUNCTION NAME:	ckvinstall
LEVEL:		3
LIBRARY:	XMODEM.LIB
DESCRIPTION:	Installs checkvalue(s) in the array of packet structures.
OTHER MODULES:	None
RETURNS:	void
COMMENTS:	numpaks is the cardinal size of the packet array. It is the
		responsibility of the error-checking function to return the
		checkvalue in the correct byte-order for storage in the
		sndpacket structure.
*/

void ckvinstall(pbp, numpaks, ckvfn)
struct	    sndpacket *pbp;	/* "packet buffer pointer"		*/
unsigned    int numpaks;	/* number of packets in buffer		*/
unsigned    int (*ckvfn)();	/* pointer to error-checking function	*/
{
    for (; numpaks > 0; --numpaks, ++pbp)
	pbp->ckval = (*ckvfn)(pbp->data);
}

/*
FUNCTION NAME:	x_snd
LEVEL:		3
LIBRARY:	XMODEM.LIB
OTHER MODULES:	From YOUR.C: rfopen, bigbuff
DESCRIPTION:	Transmits a file using XMODEM file-transfer protocol.
RETURNS:	int: 0 if successful; otherwise, the terminating error code.
*/

int x_snd(siop)
SIO *siop;
{
    extern FILE *rfopen();
    extern char *bigbuff();
    //extern unsigned short makepacket(), crchware(), *mk_crctbl();
    extern unsigned short crchware(), *mk_crctbl();
    unsigned short x_sndcksum(), x_sndcrc();

    /* static locals for speed of access    */
    static struct sndpacket *sndbuff, *sndbuffp;    /* buff & pointer	*/
    static BYTE     *bytep;	    /* for addressing structure as array*/
    static int	    ack, ackcnt, i;
    static unsigned numpax;	    /* buffer size in packets		*/
    static unsigned paxmade;	    /* number of packets received so far*/
    static unsigned short (*s_errckp)();    /* pointer to checkval funct*/
    static unsigned long paktot;    /* renning total of packets sent	*/

    FILE *sfp;
    char fnbuff[MAX_NAME_LEN];	    /* file name buffer 		*/
    int errcode;
    int errcnt;
    int paritytmp, stoptmp, dltmp;  /* to store current data format	*/
    if ((sfp = rfopen("Name for file to transmit", fnbuff)) == NIL)
	{
	printf("Cannot open %s.\n", fnbuff);	/* input file		*/
	return (1);
	}
    if ((crctblp = mk_crctbl(CRCCCITT, crchware)) == NIL) /* CRC table	*/
	{
	puts("Insufficient memory for CRC table.");
	return (1);
	}
    sndbuff = (struct sndpacket *)bigbuff(sizeof(struct sndpacket), HEADROOM, &numpax);
    if (sndbuff == NIL || numpax == 0)
	{
	puts("Insufficient memory for packet buffer.");
	return (1);
	}
    printf("Buffer size = %u packets (%lu bytes).\n", numpax, (long)numpax * CRCPAKSIZ);
    paritytmp	= paritynow(siop);  /* save current data format 	*/
    stoptmp	= stopsnow(siop);
    dltmp	= dlnow(siop);
    setparity(siop, NONE);	    /* now set for 8-N-1		*/
    setdatalen(siop, DL8);
    setstops(siop, STOP1);

    /* The following variables MUST be explicitly initialized		*/
    errcnt	= 0;	    /* maintained by x_except			*/
    paktot	= 1;	    /* running total of packets sent		*/
    paxmade	= 0;	    /* number of packets read from disk 	*/
    sndbuffp	= sndbuff;  /* pointer to main buffer			*/
    errcode	= CONTINUE; /* this value sustains the main loop	*/
    s_errckp	= NIL;	    /* use as a marker in loop below		*/

    if ((paxmade = makepacket(sndbuff, numpax, paktot, sfp)) == 0)
	errcode = x_except(siop, E_FILEMTY, &errcnt, X_ESNDMAX);
    s_clrsio(siop, CRCPAKSIZ * numpax); /* clear entire buffer		*/
    puts("Awaiting startup");
    while (errcode == CONTINUE && s_errckp == NIL)
	{
	if (inkey() == ESC)	/* user kills from keyboard		*/
	    {
	    errcode = x_except(siop, E_USRCAN, &errcnt, X_ESNDMAX);
	    break;
	    }
	switch (s_waitch(siop, _0_SEC_1))
	    {
	    case 'C':		/* receiver wants crc			*/
		s_errckp = x_sndcrc;
		paksize = CRCPAKSIZ;
		puts("\nCRC startup received"); /* now put ckvals..	*/
		ckvinstall(sndbuff, paxmade, s_errckp); /* in struct	*/
		break;
	    case NAK:		/* receiver wants checksum		*/
		s_errckp = x_sndcksum;
		paksize = CKSPAKSIZ;
		puts("\nChecksum startup received");	/* put ckvals	*/
		ckvinstall(sndbuff, paxmade, s_errckp); /* in struct	*/
		break;
	    case CAN:		/* receiver quit			*/
		errcode = x_except(siop, E_RCVCAN, &errcnt, X_ESNDMAX);
		break;
	    case TIMEOUT:	/* no timeout limit			*/
		break;
	    default:
		putchar('?');	/* not meaningful			*/
	    }
	}
    while (errcode == CONTINUE)     /* BEGIN MAIN TRANSMIT LOOP	*/
	{
	if (paxmade == 0)	    /* all packets in buffer sent?	*/
	    {
	    printf("\nDisk Read:  ");
	    sndbuffp = sndbuff;
	    paxmade = makepacket(sndbuff, numpax, paktot, sfp);
	    if (paxmade != 0)	    /* no more packets			*/
		{
		ckvinstall(sndbuff, paxmade, s_errckp);
		printf("%d packets.\n", paxmade);
		}
	    else
		{
		errcode = x_except(siop, E_EOF, &errcnt, X_ESNDMAX);
		--paktot;
		continue;
		}
	    if ferror(sfp)
		{
		errcode = x_except(siop, E_DSKREAD, &errcnt, X_ESNDMAX);
		continue;
		}
	    }
	for (i = 0, bytep = (BYTE *)sndbuffp+1; i < paksize; i++)
	    s_putc(siop, *bytep++);	/* SEND IT!			*/
	for (i = ACK_RETRY; i > 0; i--) /* get ACK or NAK		*/
	    {
	    if (inkey() == ESC) 	/* user kills from keyboard	*/
		{
		errcode = x_except(siop, E_USRCAN, &errcnt, X_ESNDMAX);
		break;
		}
	    if ((ack = s_waitch(siop, _1_SEC_0)) == ACK)
		{
		printf("%lu\r", paktot);
		++paktot;
		--paxmade;
		errcnt = 0;		/* this is optional		*/
		++sndbuffp;		/* next packet structure	*/
		break;
		}
	    if (ack == NAK || ack == 'C')	/* receiver wants last packet again	*/
		{
		errcode = x_except(siop, E_BADPAK, &errcnt, X_ESNDMAX);
		break;
		}
	    if (ack == CAN)	/* receiver wants last packet again	*/
		{
		errcode = x_except(siop, E_RCVCAN, &errcnt, X_ESNDMAX);
		break;
		}
	    }
	if (i == 0)			/* no ack received		*/
	    {
	    errcode = x_except(siop, E_NOACK, &errcnt, X_ESNDMAX);
	    s_clrsio(siop, CRCPAKSIZ);	/* clear buffer 		*/
	    }
	}		    /* END OF MAIN TRANSMIT LOOP		*/

    /* control comes here when receive loop breaks			*/
    if (errcode == E_EOF)		/* normal transfer		*/
	{
	errcode = 0;			/* for final return		*/
	printf("%lu packets sent.\nAwaiting receiver's ACK: ", paktot);
	for (i = ACK_RETRY; i > 0; i--)
	    {
	    s_putc(siop, EOT);
	    if ((ack = s_waitch(siop, _1_SEC_0 * 2)) == ACK)
		{
		putchar('\n');
		x_except(siop, E_SNDOK, &errcnt, X_ESNDMAX);
		printf(" of %s: %lu bytes.\n", fnbuff, paktot * 128);
		break;
		}
	    if (ack == CAN)		/* receiver cancelled		*/
		{
		x_except(siop, E_RCVCAN, &errcnt, X_ESNDMAX);
		break;
		}
	    if (inkey() == ESC) 	/* user killed from keyboard	*/
		{
		errcode = x_except(siop, E_USRCAN, &errcnt, X_ESNDMAX);
		ack = !ACK;		/* for the test below		*/
		break;
		}
	    putchar('.');		/* for each timeout		*/
	    }
	if (ack != ACK) 		/* final ACK never came 	*/
	    x_except(siop, E_LASTACK, &errcnt, X_ESNDMAX);
	}
    if (fclose(sfp) != 0)
	printf("\nFatal error closing %s\n", fnbuff);
    free(sndbuff);		/* release memory			*/
    free((char *)crctblp);
    setparity(siop, paritytmp); /* restore data format			*/
    setstops(siop, stoptmp);
    setdatalen(siop, dltmp);
    printf("\nEnd of X-Send\n");

    return (errcode != E_EOF);	    /* return succedd or fail		*/
}   /* end of xmit  */

/*
FUNCTION NAME:	x_rcv
LEVEL:		3
LIBRARY:	XMODEM.LIB
OTHER MODULES:	From YOUR.C wfopen, bigbuff
DESCRIPTION:	Receives a file using the XMODEM protocol.
RETURNS:	int: 0 if transfer is successful; otherwise, the terminating
		error code.
*/

int x_rcv(siop)
SIO *siop;
{
    extern unsigned short crchware(), *mk_crctbl(), x_rcvcrc();
    extern FILE *wfopen();
    extern char *bigbuff();
    extern int fclose();

    /* static locals for speed of access				*/
    static struct rcvpacket *rcvbuffp;	/* buffer pointer is global	*/
    static int	i, j;
    static short *intp; 	    /* for addressing structure as array*/
    static unsigned long paktot;    /* running packet total		*/
    static unsigned int numpax;     /* number of packets in buffer	*/
    static unsigned int buffpax;    /* number of packets received so far*/
    static int	soh;		    /* received start of header byte	*/

    FILE *rfp;			/* file pointer 			*/
    char fnbuff[MAX_NAME_LEN];	/* file name buffer			*/
    int errcode;		/* sustains main receive loop		*/
    int errcnt; 		/* main exception counter		*/
    int nosohtot;		/* special counter for soh timeouts	*/
    int paritytmp, stoptmp, dltmp;  /* to store current data format	*/

    if ((rfp = wfopen("Name of file to receive", fnbuff)) == NIL)
	{
	if (*fnbuff)
	    printf("Cannot open %s.\n", fnbuff);    /* input file	*/
	return (1);
	}
    if ((crctblp = mk_crctbl(CRCCCITT, crchware)) == NIL)   /* CRC table*/
	{
	puts("Insufficient memory for CRC table.");
	return (1);
	}
    rcvbuff = (struct rcvpacket *)bigbuff(sizeof(struct rcvpacket), HEADROOM, &numpax);
    if (rcvbuff == NIL || numpax == 0)
	{
	puts("Insufficient memory for packet buffer.");
	return (1);
	}
    printf("Buffer = %u packets (%lu bytes).\n", numpax, (long)numpax * DBLKSIZ);
    paritytmp	= paritynow(siop);	/* save current data format	*/
    stoptmp	= stopsnow(siop);
    dltmp	= dlnow(siop);
    setparity(siop, NONE);		/* now set for 8-N-1		*/
    setdatalen(siop, DL8);
    setstops(siop, STOP1);

    /* the following variables MUST be explicitly initialized		*/
    errcnt  = 0;
    buffpax = 0;		/* packets received so far		*/
    paktot  = 1;		/* running packet total 		*/
    errcode = CONTINUE;
    nosohtot = 0;		/* SOH tomeout counter			*/
    rcvbuffp = rcvbuff; 	/* pointer to main rcvbuffer		*/
    rcvbuff->pnum1 = VIRGIN;	/* used by for startup routine		*/
    r_errckp = x_rcvcrc;	/* starting check value routine 	*/
    csync   = (r_errckp == x_rcvcrc) ? 'C' : NAK;   /* match r_errckp */
    ckvname = (r_errckp == x_rcvcrc) ? "CRC" : "CKSUM";
    paksize = (r_errckp == x_rcvcrc) ? CRCPAKSIZ-1 : CKSPAKSIZ-1;

    s_clrsio(siop, CRCPAKSIZ * numpax); /* clear buffer		*/
    printf("\nAttempting to synchronize in %s mode.\n", ckvname);
    s_putc(siop, csync);	    /* send start-up sync character	*/
    while (errcode == CONTINUE)
	{
	if (inkey() == ESC)	/* user kills from keyboard		*/
	    {
	    errcode = x_except(siop, E_USRCAN, &errcnt, X_ERCVMAX);
	    continue;
	    }
	if ((soh = s_waitch(siop, SOH_TIMOUT)) != SOH)	/* bad SOH	*/
	    {
	    switch (soh)
		{
		case TIMEOUT:
		    if (!(++nosohtot % 10))	/* every 10th time	*/
			errcode = x_except(siop, E_NOSOH, &nosohtot, X_NOSOHMAX);
		    else
			s_putc(siop, csync);
		    continue;
		case EOT:	    /* no more packets			*/
		    for (rcvbuffp = rcvbuff, i = buffpax; i > 0; --i, ++rcvbuffp)
			for (j = 0; j < DBLKSIZ; j++)
			    putc(rcvbuffp->data[j], rfp);
		    if (fflush(rfp) == 0)
			{
			--paktot;
			errcode = x_except(siop, E_RCVOK, &errcnt, X_ERCVMAX);
			printf(" of %s:  %lu bytes (%lu blocks).\n", \
				fnbuff, paktot * 128, paktot);
			}
		    else
			errcode = x_except(siop, E_DSKWRITE, &errcnt, X_ERCVMAX);
		    continue;
		case CAN:	    /* sender canceled transfer 	*/
		    errcode = x_except(siop, E_SNDCAN, &errcnt, X_ERCVMAX);
		    continue;
		default:
		    errcode = x_except(siop, E_BADSOH, &errcnt, X_ERCVMAX);
		}
	    }
	for (i = 0, intp = (short *)rcvbuffp; i < paksize; i++)
	    *intp++ = s_waitch(siop, _0_SEC_1); /* rest of packet	*/
	if (rcvbuffp->ckvhi == TIMEOUT)
	    {
	    errcode = x_except(siop, E_NODATA, &errcnt, X_ERCVMAX);
	    continue;
	    }
	if ((*r_errckp)(rcvbuffp->data) != 0)	/* bad checkvalue	*/
	    {
	    errcode = x_except(siop, E_BADCKV, &errcnt, X_ERCVMAX);
	    continue;
	    }
	if (rcvbuffp->pnum1 ^ (~rcvbuffp->pnum2 & 0x00ff))  /* integrity*/
	    {
	    errcode = x_except(siop, E_PAKNUM, &errcnt, X_ERCVMAX);
	    continue;
	    }
	if (rcvbuffp->pnum1 != paktot % 256)	/* wrong sequence	*/
	    {
	    if (rcvbuffp->pnum1 == (paktot % 256) - 1)	/* previous	*/
		errcode = x_except(siop, E_SNDACK, &errcnt, X_ERCVMAX);
	    else			    /* FATAL			*/
		errcode = x_except(siop, E_PAKSEQ, &errcnt, X_ERCVMAX);
	    continue;
	    }
	printf("%lu\r", paktot);	/* running total on screen	*/
	++rcvbuffp;			/* next packet in array 	*/
	++paktot;
	errcnt = 0;
	if (++buffpax == numpax)	/* buffer full			*/
	    {
	    printf("\nDisk write\n");
	    for (rcvbuffp = rcvbuff, i = buffpax; i > 0; --i, ++rcvbuffp)
		for (j = 0; j < DBLKSIZ; j++)
		    putc(rcvbuffp->data[j], rfp);
	    if (ferror(rfp))
		{
		errcode = x_except(siop, E_DSKWRITE, &errcnt, X_ERCVMAX);
		break;
		}
	    buffpax  = 0;		/* reset counter and... 	*/
	    rcvbuffp = rcvbuff; 	/* buffer pointer		*/
	    s_clrsio(siop, CRCPAKSIZ * numpax); /* clear buffer 	*/
	    }
	s_putc(siop, ACK);		/* ACK is your final act	*/
	}
    if (fclose(rfp) != 0)   /* come here when receive loop breaks	*/
	printf("Fatal error closing %s\n", fnbuff);
    free((char *)rcvbuff);		/* release buffer		*/
    free((char *)crctblp);		/* release crc table		*/
    setparity(siop, paritytmp);
    setstops(siop, stoptmp);
    setdatalen(siop, dltmp);
    puts("\nEnd of X-Receive");
    return (errcode == E_RCVOK ? 0 : 1);    /* return succeed or fail	*/
}

/*
FUNCTION NAME:	x_sndcksum
LEVEL:		3
LIBRARY:	XMODEM.LIB
DESCRIPTION:	Calculates and returns an arithmetic checksum on the buffer
		indicated.
OTHER MODULES:	None
RETURNS:	unsigned short
*/

unsigned short x_sndcksum(buff) 	/* process checksum		*/
char *buff;				/* pointer to sndpacket->data	*/
{
    static int cksum, i;
    for (cksum = i = 0; i < 128; ++i)	/* round 'em up                 */
	cksum += *buff++;		/* add 'em up                   */
    return (cksum % 256);		/* take 'em back                */
}

/*
FUNCTION NAME:	x_rcvcksum
LEVEL:		0
LEBRARY:	...
DESCRIPTION:	Verifies 8-bit arithmetic checksum on indicated buffer.
OTHER MOKELES:	None
RETURNS:	unsigned short: 0 if checksums agree.
COMMENTS:	The checksum byte is assumed to be contiguous with the last
		byte of the data field.
*/

unsigned short x_rcvcksum(data)
int *data;			    /* pointer to rcvpacket->data	*/
{
    static int i;
    static unsigned short cksum;
    cksum = data[128];
    for (i = 0; i < DBLKSIZ; ++i)
	cksum -= *data++;	    /* subtract data bytes from checksum*/
    return (cksum % 256);
}

/*
FUNCTION NAME:	polydiv
LEVEL:		3
LIBRARY:	XMODEM.LIB
DESCRIPTION:	Simulates longhand polynomial division.
OTHER MODULES:	None
RETURNS:	unsigned short
COMMENTS:	This function produces a remainder, not a CRC.
*/

unsigned short polydiv(data, divisor, remainder)
unsigned short data;
unsigned short divisor;
unsigned short remainder;
{
    static unsigned short quotient, i;
    for (i = 8; i > 0; i--)
	{
	quotient = remainder & 0x8000;	/* remember if bit 15 is set	*/
	remainder <<= 1;		/* shift remainder; bit 0 = 0	*/
	if ((data <<= 1) & 0x0100)	/* shift data; if old bit 7 was 1   */
	    remainder |= 1;		/* put into bit 0 of remainder	*/
	if (quotient)			/* if divisible 		*/
	    remainder ^= divisor;	/* subtract polynomial		*/
	}
    return (remainder);
}

/*
FUNCTION NAME:	x_sndcrc
LEVEL		0
LIBRARY:	IBMPC.C
OTHER MODULES:	crcupdate
RETURNS:	void
COMMENTS:	This function is inherently non-portable because there is no
		way to know how the processor stores words. As shown, we
		assume that the processor stores the LSB lower in memory
		than the MSB
*/

unsigned short x_sndcrc(buff)
char *buff;			    /* pointer to sndpacket->data	*/
{
    extern void crcupdate();
    extern unsigned short *crctblp; /* pointer to lookup table		*/
    static unsigned short accum;    /* CRC accumulator			*/
    static int i;
    for (accum = i = 0; i < DBLKSIZ; ++buff, ++i)
	crcupdate(*buff, &accum, crctblp);  /* macro in CRC,H is better */
    return ((accum >> 8) + (accum << 8));   /* non-portable...		*/
}

/*
FUNCTION NAME:	x_rcvcrc
LEVEL:		3
LIBRARY:	XMODEM.LIB
DESCRIPTION:	Calculates CRC for an rcvpacket.
OTHER MODULES:	crcupdate: pointer to a CRC lookup table.
RETURNS:	unsigned short
COMMENTS:	The CRC bytes in the rcvpacket are assumed to be contiguous
		with data bytes. The CRC (in accum) is 0 if no errors occur.
*/

unsigned short x_rcvcrc(data)
short *data;			    /* pointer to rcvpacket->data	*/
{
    static unsigned short i, accum;
    for (accum = i = 0; i < DBLKSIZ+2; i++) /* include CRC bytes as data*/
	crcupdate(*data++, &accum, crctblp);
    return (accum);		/* zero if no errors			*/
}
