
#include    "stdio.h"
#include    "errno.h"
#include    "siodef.h"

/*
FUNCTION NAME:	getch
LEVEL:		3
LIBRARY:	YOUR.LIB
DESCRIPTION:	Waits until a key is pressed, then returns it without echo.
OTHER MODULES:	__keystat and __getkey low-levels.
RETURNS:	int, which allows for modification to detect "special" keys.
COMMENTS:	This should be a "raw" function with no character processing
		Lattice and Nicrosoft compilers are supported via an #ifdef
		preprocessor directive.
*/

#define NOT_READY   -1
#define NO_KEY	    0
#define MS_LATT

#ifdef MS_LATT			    /* for microsoft or Lattice 	*/
#define __keystat() _kbhit()
#define __getkey()  _getch()
#endif

int getch()			    /* fetch keyboard char, no echo	*/
{
    while (__keystat() == NO_KEY)   /* while no character is ready...	*/
	;			    /* ...wait...			*/
    return (__getkey());	    /* then return it when ready	*/
}

/*
FUNCTION NAME:	inkey
LEVEL:		3
LIBRARY:	YOUR.LIB
DESCRIPTION:	Returns a keyboard character if available; otherwise -1.
OTHER MODULES:	__keystat, __getkey
RETURNS:	int, -1 if no character ready.
COMMENTS:	This should be a "raw" function with no character processing
		Lattice and Nicrosoft compilers are supported via an #ifdef
		preprocessor directive.
*/

int inkey()
{
    return ((__keystat() == NO_KEY) ? NOT_READY : __getkey());
}

/*
FUNCTION NAME:	bigbuff
LEVEL:		3
LIBRARY:	YOUR.C
DESCRIPTION:	Allocates large buffer.
OTHER MODULES:	None
RETURNS:	char *: the address of the allocated buffer or NIL if no
		memory is available.
COMMENTS:	Alllocates nblks of memory, in blksize chunks. The hesdroom
		argument is the number of blksize chunks to return to the
		system for local use. The actual number of blocks allocated
		is returned to the caller via the pointer numblkp.
*/

//#include    "stdio.h"
//#include    "errno.h"
//#include    "siodef.h"

char *bigbuff(blksize, headroom, numblkp)
unsigned blksize;	    /* block size for allocation		*/
unsigned headroom;	    /* number of blocks NOT to allocate 	*/
unsigned *numblkp;	    /* pointer caller's count variable          */
{
    extern char *malloc();
    char *p;
    unsigned nblks = 1;
    while ((p = malloc(nblks * blksize)) != NIL)    /* call until failure */
	{
	free(p);		/* release it				*/
	++nblks;		/* ask for more next time		*/
	}
    if (headroom >= --nblks)	/* not enough memory			*/
	{
	nblks = 0;
	return (NIL);
	}
    nblks -= headroom;		/* adjust for headroom			*/
    *numblkp = nblks;		/* pass back to caller			*/
    return (malloc(nblks * blksize));	/* allocate & return the pointer*/
}

/*
FUNCTION NAME:	putbin
LEVEL:		3
LIBRARY:	YOUR.LIB
DESCRIPTION:	Desplays a 16-bit integer in binary.
RETURNS:	void
COMMENTS:	Newline at the end is to accommodate the program in figure
		19-3.
*/

void putbin(data)
unsigned short data;		    /* 16-bit int			*/
{
    unsigned short shifter = 0x8000;
    for (; shifter > 0; shifter >>= 1)
	putchar((data & shifter) ? '1' : '0');
    putchar('\n');
}

/*
FUNCTION NAME:	rfopen
LEVEL:		3
LIBRARY:	YOUR.LIB
DESCRIPTION:	Continually prompts for filename until a file is opened for
		reading or the user aborts.
OTHER MODULES:	None
RETURNS:	FILE *: the FILE pointer or NIL if the open is unsuccessful.
COMMENTS:	The filename buffer passed must be long enough to hold the
		largest valid system name (79 bytes in PC DOS, 14 in CP/M).
*/

//#include    "stdio.h"
//#include    "errno.h"

FILE *rfopen(promptstr, fnbuff)
char *promptstr;	      /* user prompt string		      */
char *fnbuff;		      /* file name buffer		      */
{
    //extern int errno;   /* this constant has declared in error.h    */
    extern FILE *fopen();
    FILE *rfp;
    int c;
    int errflag = 0;
    do	      {
      printf("\n%s: ", promptstr);
      gets(fnbuff);
      if (strcmp(fnbuff, "") == 0)    /* user typed CR to exit	      */
	  return (NIL);
      if ((rfp = fopen(fnbuff, "rb")) == NIL)  /* open failure	      */
	  if (errno == ENOENT)
	      printf("%s not found.", fnbuff);
	  else
	      printf("\nOpen error on %s.", fnbuff);  /* error message*/
      } while (rfp == NIL);	  /* repeat until open succeds	      */
    return (rfp);
}

/*
FUNCTION NAME:	substr
LEVEL:		3
LIBRARY:	YOUR.LIB
DESCRIPTION:	Ascertains if one string is a substring of the other.
OTHER MODULES:	index (strchr on some compilers)
RETURNS:	char *: pointer to the first character of minstr within
		majstr; otherwise, NIL.
COMMENTS:	An "ignore case" mode would be useful.
*/

//#include    "stdio.h"

char *substr(majstr, minstr)
char *majstr;			/* big string				*/
char *minstr;			/* little string			*/
{
    extern char *index();
    char *gp = majstr;
    int sublen;
    if ((sublen = strlen(minstr)) > strlen(majstr))
	return (NIL);		/* substring shorter than main string	*/
    while ((gp = index(gp, *minstr)) != NIL)	/* find match of first char*/
	{
	if (strncmp(gp, minstr, sublen) == NUL)
	    return (gp);
	++gp;
	}
    return (NIL);
}

/*
FUNCTION NAME:	wfopen
LEVEL:		3
LIBRARY:	YOUR.LIB
DESCRIPTION:	Continually prompts for the name of an output file until
		the file is successfully opened or the user aborts.
OTHER MODULES:	None
RETURNS:	FILE *: the FILE pointer or NIL if the open is unsuccessful.
COMMENTS:	Asks permission before overwriting a pre-existing file.
		The filename buffer passed must be long enough to hold the
		largest valid system name (79 bytes in PC DOS, 14 in CP/M).
*/

//#include    "stdio.h"
//#include    "errno.h"     /* maybe just "error.h" */

FILE *wfopen(promptstr, fnbuff)
char *promptstr;
char *fnbuff;
{
    FILE *wfp;
    char locbuff[10];
    for (EVER)
	{
	printf("\n%s: ", promptstr);
	gets(fnbuff);
	if (strcmp(fnbuff, "") == 0)	    /* new line typed		*/
	    return (NIL);
	fclose(wfp = fopen(fnbuff, "rb"));   /* open (and close) it for reading	*/
	if (wfp == NIL) 		    /* doesn't already exist    */
	    return (fopen(fnbuff, "wb"));
	printf("\n%s already exists. Overwrite\007? ", fnbuff);
	gets(locbuff);
	if (toupper(locbuff[0]) == 'Y')
	    break;			    /* ok to overwrite		*/
	}
    return (fopen(fnbuff, "wb"));
}
